#!/bin/env python

# -----------------------------------------------------------------------------
# HMM simulation functions
# Author: Mary Richardson
# Date: 2020.01.08
# -----------------------------------------------------------------------------

import logging
import numpy as np

logger = logging.getLogger(__name__)

from orfeus.hmm.helper import convert_emissions
from orfeus.hmm.algorithms import viterbi


def sample_transitions(p_trans):
    """
    Sample a path from the model, using the model transitions
    Args:
        p_emit_nt: the nucleotide emission probability matrix
        p_emit_rho: the riboseq emission probability matrix
        p_trans: the transition probability matrix
        nts: list of valid nucleotides
        rho_ranges: array of riboseq emission bin ranges
    Returns:
        path: simulated path
    Raises:

    """
    M = len(p_trans) # Number of states

    # Choose a state path using transition probabilities
    path = []

    # Start at the begin state and correct for the offset (1-indexed)
    next_state = np.random.choice(M, p=p_trans[0]) - 1

    # Continue adding next states until you reach the end state
    while next_state != -1:
        curr_state = next_state
        path.append(curr_state)
        next_state = np.random.choice(M, p=p_trans[curr_state+1]) - 1

    return path


def sample_emissions(path,
                 p_emit_nt,
                 p_emit_rho,
                 nts,
                 rho_ranges):
    """
    Sample a sequence for the path from the model, using the model emissions
    Args:
        path: simulated path
        p_emit_nt: the nucleotide emission probability matrix
        p_emit_rho: the riboseq emission probability matrix
        nts: list of valid nucleotides
        rho_ranges: array of riboseq emission bin ranges
    Returns:
        x: simulated nt sequence
        r: simulated rho sequence
    Raises:

    """
    # Set possible emission values (zero if zero bin, mean of bin otherwise)
    rhos = np.append([0], np.mean([rho_ranges[1:-1],
                                   rho_ranges[2:]],
                                   axis=0))

    # Choose the nucleotide and riboseq observation for each state using
    # emission probabilities
    x = []
    r = []
    for s in path:

        # Assign the nt value
        x.append(np.random.choice(nts, p=p_emit_nt[s]))

        # Assign the rho value
        r.append(np.random.choice(rhos, p=p_emit_rho[s]))

    x = np.array(x)
    r = np.array(r)

    # Normalize rho by mean rho over the simulated sequence
    r = r/np.mean(r)

    return x, r


def simulate_seq_any(p_emit_nt,
                     p_emit_rho,
                     p_trans,
                     nts,
                     rho_ranges):
    """
    Sample a sequence from the model (can be canonical or noncanonical)
    Args:
        p_emit_nt: the nucleotide emission probability matrix
        p_emit_rho: the riboseq emission probability matrix
        p_trans: the transition probability matrix
        nts: list of valid nucleotides
        rho_ranges: array of riboseq emission bin ranges
    Returns:
        x: simulated nt sequence
        r: simulated rho sequence
        path: simulated path
    Raises:

    """
    # Sample a path using the transitions
    path = sample_transitions(p_trans)

    # Sample a nt and rho sequence using the emissions
    x, r = sample_emissions(path,
                            p_emit_nt,
                            p_emit_rho,
                            nts,
                            rho_ranges)

    return x, r, path


def simulate_seq_canonical(p_emit_nt,
                           p_emit_rho,
                           p_trans,
                           nts,
                           rho_ranges,
                           altorfs_idx):
    """
    Sample a canonical sequence from the model
    Args:
        p_emit_nt: the nucleotide emission probability matrix
        p_emit_rho: the riboseq emission probability matrix
        p_trans: the transition probability matrix
        nts: list of valid nucleotides
        rho_ranges: array of riboseq emission bin ranges
        altorfs_idx: dictionary mapping more general state type to
            state indices (0-indexed)
    Returns:
        x: simulated nt sequence
        r: simulated rho sequence
        path: simulated path
    Raises:

    """
    altorfs = altorfs_idx['uORF'] + \
              altorfs_idx['dORF'] + \
              altorfs_idx['PRF'] + \
              altorfs_idx['SCR']

    # If the simulated sequence contains non-canonical states,
    # regenerate until it doesn't
    noncanonical = True
    while noncanonical:

        # Sample a path using the transitions
        path = sample_transitions(p_trans)

        # Check if non-canonical
        noncanonical = any([path[i] in altorfs for i in range(len(path))])

    # Sample a nt and rho sequence using the emissions
    x, r = sample_emissions(path,
                            p_emit_nt,
                            p_emit_rho,
                            nts,
                            rho_ranges)

    return x, r, path


def check_unique(path,
                 altorfs_idx,
                 event):
    """
    Determine whether there is more than one simulated altORF in a path
    Args:
        path: simulated path
        altorfs_idx: dictionary mapping more general state type to
            state indices (0-indexed)
        event: altORF event type to simulate
            uORF = uORF only
            dORF = dORF only
            uORFdORF = uORF or dORF
            pPRF = +1 frameshift
            mPRF = -1 frameshift
            SCR = stop codon readthrough
    Returns:
        unique: boolean indicating whether there is only one altORF
    Raises:

    """
    # Check that the altORF event is only one of its kind in the path
    unique = True
    if event=='uORF':
        if len([i for i in path if i in altorfs_idx['start_uORF']]) > 3:
            unique = False
    elif event=='dORF':
        if len([i for i in path if i in altorfs_idx['start_dORF']]) > 3:
            unique = False
    elif event=='uORFdORF':
        if len([i for i in path if i in (altorfs_idx['start_uORF'] + \
                                         altorfs_idx['start_dORF'])]) > 3:
            unique = False
    elif event=='pPRF':
        if len([i for i in path if i==altorfs_idx['X1']]) > 1:
            unique = False
    elif event=='mPRF':
        if len([i for i in path if i==altorfs_idx['X2']]) > 1:
            unique = False
    elif event=='SCR':
        if len([i for i in path if i in altorfs_idx['SCR']]) > 3:
            unique = False

    return unique


def simulate_seq_noncanonical(p_emit_nt,
                              p_emit_rho,
                              p_trans,
                              p_trans_rev,
                              nts,
                              rho_ranges,
                              altorfs_idx,
                              event):
    """
    Sample a noncanonical sequence from the model
    Args:
        p_emit_nt: the nucleotide emission probability matrix
        p_emit_rho: the riboseq emission probability matrix
        p_trans: the transition probability matrix
        p_trans_rev: the reverse transition probability matrix
        nts: list of valid nucleotides
        rho_ranges: array of riboseq emission bin ranges
        altorfs_idx: dictionary mapping more general state type to
            state indices (0-indexed)
        event: altORF event type to simulate
            uORF = uORF only
            dORF = dORF only
            uORFdORF = uORF or dORF
            pPRF = +1 frameshift
            mPRF = -1 frameshift
            SCR = stop codon readthrough
    Returns:
        x: simulated nt sequence
        r: simulated rho sequence
        path: simulated path
    Raises:

    """
    M = len(p_trans) # Number of states

    # Choose a state path using transition probabilities
    # Start at the noncanonical state
    if event=='uORF':
        idx = np.random.choice(altorfs_idx['start_uORF'])
    elif event=='dORF':
        idx = np.random.choice(altorfs_idx['start_dORF'])
    elif event=='uORFdORF':
        idx = np.random.choice(altorfs_idx['start_uORF'] + \
                               altorfs_idx['start_dORF'])
    elif event=='pPRF':
        idx = altorfs_idx['X1']
    elif event=='mPRF':
        idx = altorfs_idx['X2']
    elif event=='SCR':
        idx = np.random.choice(altorfs_idx['SCR'])
    else:
        logger.error(('Invalid event type specified for simulating sequence. '
                      'Choose from: uORF, dORF, uORFdORF, pPRF, mPRF, SCR'))
    path = [idx]

    # Add the next state and check
    next_state = np.random.choice(M, p=p_trans[idx+1]) - 1
    # For +PRF make sure the next state is not a -PRF
    if event=='pPRF':
        while next_state == altorfs_idx['X2']:
            next_state = np.random.choice(M, p=p_trans[idx+1]) - 1

    # Continue adding next states until you reach the end state
    while next_state != -1:
        curr_state = next_state
        path.append(curr_state)
        next_state = np.random.choice(M, p=p_trans[curr_state+1]) - 1

    # Continue adding previous states until you reach the begin state
    prev_state = np.random.choice(M, p=p_trans_rev[idx+1]) - 1
    while prev_state != -1:
        curr_state = prev_state
        path.insert(0, prev_state)
        prev_state = np.random.choice(M, p=p_trans_rev[curr_state+1]) - 1

    # Check that the altORF is only one of its kind in the simulated sequence
    # If not, resimulate until it is
    unique = check_unique(path, altorfs_idx, event)
    if not unique:
        return simulate_seq_noncanonical(p_emit_nt,
                                         p_emit_rho,
                                         p_trans,
                                         p_trans_rev,
                                         nts,
                                         rho_ranges,
                                         altorfs_idx,
                                         event)

    # Sample a nt and rho sequence using the emissions
    x, r = sample_emissions(path,
                            p_emit_nt,
                            p_emit_rho,
                            nts,
                            rho_ranges)

    return x, r, path


def scale_coverage(r, coverage):
    """
    Scale riboseq read coverage to reach target coverage (reads/nt)
    Args:
        r: simulated rho sequence
        coverage: mean transcript coverage to simulate (reads/nt)
    Returns:
        r: updated simulated rho sequence
    Raises:

    """
    # Scale the coverage as necessary
    if coverage > 0:

        # Convert riboseq values to a probability distribution
        r_prob = r/sum(r)

        # Calculate how many reads to assign to the transcript to generate
        # the target coverage
        reads = int(coverage * len(r))

        # Assign reads to the simulated sequence based on the probability
        # distribution and the target coverage
        L = len(r)
        r = np.zeros(L)
        for _ in range(reads):
            i = np.random.choice(L, p=r_prob)
            r[i] += 1

        # Renormalize the reads
        if np.mean(r) == 0: r = np.zeros(L)
        else: r = r/np.mean(r)

    else:
        logger.warning(('Cannot scale coverage for sequence. '
                        'Target coverage is less than or equal to zero.'))

    return r


def simulate_seq(p_emit_nt,
                 p_emit_rho,
                 p_trans,
                 p_trans_rev,
                 nts,
                 rho_ranges,
                 altorfs_idx,
                 event,
                 coverage=0):
    """
    Sample a sequence of the given event type from the model
    Args:
        p_emit_nt: the nucleotide emission probability matrix
        p_emit_rho: the riboseq emission probability matrix
        p_trans: the transition probability matrix
        p_trans_rev: the reverse transition probability matrix
        nts: list of valid nucleotides
        rho_ranges: array of riboseq emission bin ranges
        altorfs_idx: dictionary mapping more general state type to
            state indices (0-indexed)
        event: event type to simulate
            any = any ORF(s), canonical or noncanonical
            ORF = canonical ORF
            uORF = uORF only
            dORF = dORF only
            uORFdORF = uORF or dORF
            pPRF = +1 frameshift
            mPRF = -1 frameshift
            SCR = stop codon readthrough
        coverage: (optional) mean transcript coverage to simulate (reads/nt),
            default is not to scale the rho values at all
    Returns:
        x: simulated nt sequence
        r: simulated rho sequence
        path: simulated path
    Raises:

    """
    # Simulate a sequence
    if event=='any':
        x, r, path = simulate_seq_any(p_emit_nt,
                                      p_emit_rho,
                                      p_trans,
                                      p_trans_rev,
                                      nts,
                                      rho_ranges)
    elif event=='ORF':
        x, r, path = simulate_seq_canonical(p_emit_nt,
                                            p_emit_rho,
                                            p_trans,
                                            p_trans_rev,
                                            nts,
                                            rho_ranges,
                                            altorfs_idx)

    elif event in ['uORF','dORF','uORFdORF','pPRF','mPRF','SCR']:
        x, r, path = simulate_seq_noncanonical(p_emit_nt,
                                               p_emit_rho,
                                               p_trans,
                                               p_trans_rev,
                                               nts,
                                               rho_ranges,
                                               altorfs_idx,
                                               event)

    else:
        logger.error(('Invalid event type specified for simulating sequence. '
                      'Choose from: uORF, dORF, pPRF, mPRF, SCR, any, ORF'))

    # If there are no reads, run again recursively
    if (sum(r)==0):
        logger.warning('No reads simulated. Simulating new sequence...')
        return simulate_seq(p_emit_nt,
                            p_emit_rho,
                            p_trans,
                            p_trans_rev,
                            nts,
                            rho_ranges,
                            altorfs_idx,
                            event,
                            coverage)

    # Scale coverage to reach target coverage
    r = scale_coverage(r, coverage)

    return x, r, path


def simulation(p_emit_nt,
               p_emit_rho,
               p_trans,
               p_trans_rev,
               log_emit_nt,
               log_emit_rho,
               log_trans,
               prev_states_options,
               next_states_options,
               nts,
               rho_ranges,
               altorfs_idx,
               coverage,
               event='any'):
    """
    Simulates a path and corresponding sequence of nts and rho values,
    then runs Viterbi on the sequence to predict the path
    Args:
        p_emit_nt: the nucleotide emission probability matrix
        p_emit_rho: the riboseq emission probability matrix
        p_trans: the transition probability matrix
        p_trans_rev: the reverse transition probability matrix
        log_emit_nt: the nucleotide log emission probability matrix
        log_emit_rho: the riboseq log emission probability matrix
        log_trans: the log transition probability matrix
        prev_state_options: list of indices of all possible
            (nonzero probability) previous states
        next_state_options: list of indices of all possible
            (nonzero probability) next states
        nts: list of valid nucleotides
        rho_ranges: array of riboseq emission bin ranges
        altorfs_idx: dictionary mapping more general state type to
            state indices (0-indexed)
        coverage: riboseq coverage (mean reads/nt)
        event: (optional) event type to look for
            ORF = canonical ORF
            uORF = uORF only
            dORF = dORF only
            uORFdORF = uORF or dORF
            pPRF = +1 frameshift
            mPRF = -1 frameshift
            SCR = stop codon readthrough
            default is any event
    Returns:
        true_path: simulated path
        pred_path: predicted path
    Raises:

    """
    # Simulate sequence
    x, r, true_path = simulate_seq(p_emit_nt,
                                   p_emit_rho,
                                   p_trans,
                                   p_trans_rev,
                                   nts,
                                   rho_ranges,
                                   altorfs_idx,
                                   event,
                                   coverage)

    # Generate the emission probabilities for this sequence
    log_emit = convert_emissions(x,
                                 r,
                                 log_emit_nt,
                                 log_emit_rho,
                                 nts,
                                 rho_ranges)

    # Predict the viterbi state path
    pred_path, P = viterbi(log_emit,
                           log_trans,
                           prev_states_options,
                           next_states_options)

    return true_path, pred_path



#!/usr/bin/env python

# -----------------------------------------------------------------------------
# Hidden Markov Model Algorithms
# Author: Mary Richardson
# Date: 2020.01.08
# -----------------------------------------------------------------------------

import logging
import numpy as np
from scipy.special import logsumexp

logger = logging.getLogger(__name__)

# Ignore divide by zero error for log
np.seterr(divide='ignore')


def viterbi(log_emit,
            log_trans,
            prev_states_options,
            next_states_options):
    """
    Determine the best path for a sequence using the Viterbi algorithm
    A note about dimensions:
        transition matrix includes begin state (1-indexed)
        next and prev state options include begin state (1-indexed)
        emission matrix excludes begin state (0-indexed)
        viterbi matrix excludes begin state (0-indexed)
    Args:
        log_emit: the total emission probabilities at each position for
            this sequence (combined nucleotide and rho emissions)
        log_trans: log transition probability matrix
        prev_state_options: list of indices of all possible
            (nonzero probability) previous states
        next_state_options: list of indices of all possible
            (nonzero probability) next states
    Returns:
        path: Viterbi path
        P: total probability of Viterbi path
    Raises:

    """
    L = len(log_emit)                   # Length of observed sequence
    M = len(log_trans) - 1              # Number of states
    V = np.log(np.zeros([L,M]))         # Viterbi score matrix
    ptr = np.zeros([L,M])               # Viterbi traceback matrix
    path = np.zeros(L, dtype=np.int32)  # Most probable viterbi path

    # Initialization

    # All sequences start in the begin state
    for k in states:
        V[0,k-1] = log_emit[0,k-1] + log_trans[0,k]

    # Recursion

    # For each position in the observed sequence 1..L
    for i in range(1, L):

        # For each possible next state
        for k in states:
            l = state_length[k]
            Z = [V[i-l,j-1] + log_trans[j,k] for j in states]

            # Update the viterbi score matrix
            V[i,k-1] = log_emit[[i-l+1:i],k-1] + np.max(Z)

            # Update the viterbi traceback matrix
            ptr[i,k-1] = states[np.argmax(Z)] - 1

    # Termination

    # All sequences end in the end state
    Z = [V[L-1,j-1] + log_trans[j,0] for j in states]

    # Total probability of sequence
    P = np.max(Z)

    # Traceback

    # End in the end state
    path[L-1] = states[np.argmax(Z)] - 1

    # For each position in the observed sequence L-1..0
    for i in reversed(range(L-1)):

        # Traceback to the most probable previous state FIXME!
        path[i] = ptr[i+1, path[i+1]]

    return path, P


def score_path(path,
               log_emit,
               log_trans):
    """
    Calculates the score of the given state path, using the given model
    parameters
    Args:
        path: state path
        log_emit: the total emission probabilities at each position for
            this sequence (combined nucleotide and rho emissions)
        log_trans: log transition probability matrix
    Returns:
        P: total probability of the path
    Raises:

    """
    # Length of observed sequence
    L = len(path)

    if L==0:
        return np.nan

    # Add the probability of the initial transition
    # from the begin state (in log space)
    P = log_trans[0, path[0]+1]

    # For each position in the sequence 0..L-1
    # add the probability of the transition to this residue and
    # the emission for this residue (in log space)
    for i in range(0, L-1):
        P += log_emit[i, path[i]] + log_trans[path[i]+1, path[i+1]+1]

    # Add the probability of the final emission and
    # the transition to the end state (in log space)
    P += log_emit[L-1, path[L-1]] + log_trans[path[L-1]+1, 0]

    return P



#!/usr/bin/env python

# -----------------------------------------------------------------------------
# Coverage threshold
# Author: Mary Richardson
# Date: 2020.06.04
# -----------------------------------------------------------------------------

from orfeus.coverage.sensitivity import sensitivity_simulation, plot_sensitivity
from orfeus.coverage.specificity import specificity_simulation, plot_specificity
from orfeus.coverage.accuracy import accuracy_simulation, plot_accuracy
from orfeus.coverage.count import plot_gene_coverage

NUCLEOTIDES = ['A','C','G','T']



def coverage_threshold_simulation(data_df,
                                  parameters,
                                  coverages,
                                  iters,
                                  window,
                                  outdir):

    # For each event type
    events = ['uORFdORF','pPRF','mPRF','SCR','ORF']
    for event in events:

        # Sensitivity simulation
        sensitivity = sensitivity_simulation(parameters,
                                             NUCLEOTIDES,
                                             coverages,
                                             iters,
                                             window,
                                             event)

        # Specificity simulation
        specificity = specificity_simulation(parameters,
                                             NUCLEOTIDES,
                                             coverages,
                                             iters,
                                             event)

        # Plot results
        sens_file = os.path.join(outdir, 'sensitivity_' + event + '.pdf')
        spec_file = os.path.join(outdir, 'specificity_' + event + '.pdf')
        plot_sensitivity(sensitivity, event, sens_file)
        plot_specificity(specificity, event, spec_file)

    # Accuracy simulation
    accuracy = accuracy_simulation(parameters,
                                   NUCLEOTIDES,
                                   coverages,
                                   iters)

    # Plot results
    acc_file = os.path.join(outdir, 'accuracy.pdf')
    cov_file = os.path.join(outdir, 'transcripts.pdf')
    plot_accuracy(accuracy, acc_file)
    plot_gene_coverage(data_df, coverages, cov_file)



#!/usr/bin/env python

# -----------------------------------------------------------------------------
# Data processing
# Author: Mary Richardson
# Date: 2020.01.08, last update 2023.03.01
# -----------------------------------------------------------------------------

import logging
import numpy as np
import pandas as pd

import orfeus.parsers.wig as wig
import orfeus.parsers.bg as bg
import orfeus.parsers.fa as fa
import orfeus.parsers.gtf as gtf

logger = logging.getLogger(__name__)

# Ignore divide by zero error for log
np.seterr(divide='ignore')



def read_counts_data(file, strand):
    """
    Generate a DataFrame containing the ribo-seq read counts at each
    position of the genome
    Args:
        file: path to a read counts Wiggle or BedGraph file
        strand: strand of reads specified in the input file ('+' or '-)'
    Returns:
        df: pandas DataFrame with the read counts and strand
    Raises:
        IOError: an error occurred accessing the read counts file
        IndexError, TypeError, KeyError: an error occurred reading the
            read counts file
    """
    # Parse Wiggle or BedGraph file
    try: # Try to parse as wiggle
        df = wig.wig_to_df(file)
    except:
        try: # Try to parse as bedgraph
            df =  bg.bg_to_df(file)
        except:
            logging.error(('Could not parse read counts file %s. '
                    'Make sure file is a Wiggle (.wig) or BedGraph (.bg) file.')
                    % file)
            raise

    df['strand'] = strand
    return df


def seq_data(file, strand):
    """
    Generate a DataFrame containing the nucleotide sequence at each
    position of the genome or transcriptome
    Args:
        file: path to a fasta file
        strand: strand of seqs specified in the input file ('+' or '-)'
    Returns:
        df: pandas DataFrame with the seq and strand
    Raises:
        IOError: an error occurred accessing the seq file
        IndexError, TypeError, KeyError: an error occurred reading the
            seq counts file
    """
    # Parse fasta file
    try:
        if strand=='+':
            df = fa.fasta_to_df(file, rev_comp=False)
        elif strand=='-':
            df = fa.fasta_to_df(file, rev_comp=True)

    except:
        logging.error(('Could not parse read counts file %s. '
                'Make sure file is a Fasta file.')
                % file)
        raise

    df['strand'] = strand
    return df


def feature_label(feature):
    """
    Updates feature labels to shortened feature names
    Args:
        feature: current feature name
    Returns:
        feature: updated feature name
    """
    if feature=='five_prime_utr':
            return '5UTR'
    elif feature=='exon':
            return 'ORF'
    elif feature=='three_prime_utr':
            return '3UTR'
    else: return feature


def annotations_data(file):
    """
    Parse a GTF/GFF file to extract all annotated transcripts
    Args:
        file: path to an annotations gff/gtf file
    Returns:
        df: pandas DataFrame of all annotated transcripts
    Raises:
        IOError: an error occurred accessing the file
    """
    # Parse input info
    try:
        df = gtf.gtf_to_df(file)
    except:
        logging.error(('Could not parse annotations file %s. '
                'Make sure file is a GTF or GFF file.')
                % file)
        raise

    # Filter down to columns of interest
    df = df[['chrom',
             'transcript_id',
             'transcript_name',
             'transcript_biotype',
             'feature',
             'strand',
             'start',
             'end']]

    # Filter down to features of interest (protein-coding UTRs and ORFs)
    df = df[df['transcript_biotype'].isin(['protein_coding'])]
    df = df[df['feature'].isin(['five_prime_utr',
                                'exon',
                                'three_prime_utr'])]
    df['feature'] = pd.Categorical(df['feature'].map(feature_label),
                                   categories=['5UTR', 'ORF', '3UTR'],
                                   ordered=True)

    # Sort by chrom then transcript then feature
    df.sort_values(by=['chrom',
                       'transcript_id',
                       'feature'],
                   inplace=True)

    # Add chromosome position
    df['chrom_pos'] = df.apply(lambda x: list(range(int(x['start']),
                                                   int(x['end'])+1)),
                                                   1)
    df = df.explode('chrom_pos')
    df = df.drop(columns=['start', 'end'])

    # Add transcript position
    df['transcript_pos'] = df.groupby('transcript_id') \
                             .cumcount()+1

    # Add feature position
    df['feature_pos'] = df.groupby(['transcript_id','feature']) \
                          .cumcount()+1

    # Add transcript length
    df['transcript_len'] = df.groupby(['transcript_id']) \
                                      ['transcript_pos'].transform(len)

    # Add feature length
    df['feature_len'] = df.groupby(['transcript_id','feature']) \
                                   ['feature_pos'].transform(len)

    if df.empty:
        logging.error(('Processed data is empty. Check formatting of GTF/GFF'
                'file %s') % file)
        raise

    return df


def process_data(plus_file,
                 minus_file,
                 seqs_file,
                 annotations_file,
                 skip=[]):
    """
    Merge riboseq data and annotations data into one DataFrame
    Args:
        plus_file: plus strand read counts bg/wig file
        minus_file: minus strand read counts bg/wig file
        seqs_file: genome or transcriptome nucleotide sequence fasta file
        annotations_file: annotations gff/gtf file
        skip: (optional) names of frameshifted genes to exclude
            when adding codon/state labels (recommended to prevent out-of-frame
            and partial codons from being added)
    Returns:
        data_df: pandas DataFrame of all annotated transcripts
    Raises:
        IOError: an error occurred accessing the file
        ValueError: could not merge sequence and annotation data
    """
    # Annotations
    annot_df = annotations_data(annotations_file)

    # Read counts
    reads_plus_df = read_counts_data(plus_file, '+')
    reads_minus_df = read_counts_data(minus_file, '-')
    reads_df = pd.concat([reads_plus_df, reads_minus_df], axis=0)

    # Sequences
    seqs_plus_df = seq_data(seqs_file, '+')
    seqs_minus_df = seq_data(seqs_file, '-')
    seqs_df = pd.concat([seqs_plus_df, seqs_minus_df], axis=0)

    # Merge read counts and annotations
    df = pd.merge(annot_df, reads_df,
                  on=['chrom','chrom_pos','strand'],
                  how='left')

    # Fill any missing read read_counts positions with zero
    df['reads'].fillna(0, inplace=True)

    # Merge sequence by chromosomes
    if set(seqs_df['seq_id'].unique()) == set(annot_df['chrom'].unique()):

        # Add chromosome length
        chrom_len = seqs_df['seq_id'].value_counts().to_dict()
        df['chrom_len'] = df['chrom'].map(chrom_len)//2

        # Correct the genome position to account for strand
        df['chrom_pos'] = np.where(df['strand']=='-', \
                          (df['chrom_len']-df['chrom_pos']+1), \
                          df['chrom_pos'])

        # Correct the transcript position to account for strand
        df['feature_pos'] = np.where(df['strand']=='-', \
                            (df['feature_len']-df['feature_pos']+1), \
                            df['feature_pos'])

        # Re-sort by corrected feature position
        df.sort_values(by=['chrom',
                           'transcript_id',
                           'feature',
                           'feature_pos'],
                       inplace=True)

        # Correct transcript position
        df['transcript_pos'] = df.groupby('transcript_id') \
                                 .cumcount()+1

        df = pd.merge(df, seqs_df,
                      left_on=['chrom','chrom_pos','strand'],
                      right_on=['seq_id','seq_pos','strand'],
                      how='left') \
               .drop(columns=['seq_id','seq_pos'])

    # Or merge sequence by transcripts
    elif set(seqs_df['seq_id'].unique()) == set(annot_df['transcript_id'].unique()):
        df = pd.merge(df, seqs_df,
                      left_on=['transcript_id','transcript_pos','strand'],
                      right_on=['seq_id','seq_pos','strand'],
                      how='left') \
               .drop(columns=['seq_id','seq_pos'])

    else:
        logging.error(('Could not merge sequence and annotation data. '
                'Make sure sequence IDs (chromosomes or transcripts) are the '
                'same in the sequence and annotation files.'))
        raise

    # Add nt position (for ORF only)
    df['nt_pos'] = np.where(df['feature'] == 'ORF',
                           ((df['feature_pos']+2)%3+1),
                           '')

    # Add codon position (for ORF only)
    df['codon_pos'] = np.where(df['feature'] == 'ORF',
                              (df['feature_pos']-1)//3,
                              -1)

    # Add codon type (for ORF only)
    conditions = [(df['feature'] == 'ORF') & \
                  (df['codon_pos'] == 0),
                  (df['feature'] == 'ORF') & \
                  (df['codon_pos'] == df['feature_len']//3-1)]
    choices = ['START', 'STOP']
    df['codon_type'] = np.select(conditions,
                                 choices,
                                 default=df['feature'])

    # Normalize riboseq reads by the mean riboseq reads per transcript
    df['mean_reads'] = df.groupby('transcript_id') \
                                   ['reads'].transform('mean')
    #df['mean_orf_reads'] = df.groupby(['gene_id','feature']) \
    #                                    ['reads'].transform('mean')
    #df.loc[data_df['feature']!='ORF', 'mean_orf_reads'] = np.nan
    #df['mean_orf_reads'] = df.groupby('gene_id') \
    #                                    ['mean_orf_reads'].transform('first')

    # Normalize by mean reads over transcript
    df['norm_reads'] = (df['reads']/df['mean_reads']) \
                                .replace(np.nan, 0)
    df['log_reads'] = np.log(df['norm_reads'] + 1)

    # Find the codon sequence at each genome position that contains a ORF
    codons_df = df[df['feature']=='ORF'].copy()

    # Ignore known frameshifted genes,
    # since the codon sequence will be incorrectly inferred
    codons_df = codons_df[~codons_df['transcript_name'].isin(skip)]

    # Codons are just every three nucleotides
    codons_df['codon_seq'] = codons_df.groupby(['transcript_id','codon_pos']) \
                                      ['seq'].transform(lambda x: ''.join(x))

    # Add the codon sequence back to the final dataframe
    df = df.merge(codons_df[['transcript_id',
                             'transcript_pos',
                             'codon_seq']],
                  on=['transcript_id','transcript_pos'],
                  how='left')

    # Delete old dataframe
    del codons_df

    # Add the state to the dataframe
    conditions = [(df['codon_type'] == 'START'),
                  (df['codon_type'] == 'STOP'),
                  (df['codon_type'] == 'ORF'),
                  (df['feature'] == '5UTR') | (df['feature'] == '3UTR')]
    choices = [df['codon_seq'] + df['nt_pos'].astype(str) + '#',
               df['codon_seq'] + df['nt_pos'].astype(str) + '*',
               df['codon_seq'] + df['nt_pos'].astype(str),
               df['feature']]
    df['state'] = np.select(conditions, choices, default='')

    # Clean up the columns
    df.drop(columns=['chrom_pos', 'codon_pos', 'chrom_len'],
                     inplace=True)
    dtype0 = {'transcript_id': 'str',
             'transcript_name': 'str',
             'transcript_pos': 'int64',
             'transcript_len': 'int64',
             'feature': 'str',
             'feature_pos': 'int64',
             'feature_len': 'int64',
             'seq': 'str',
             'codon_seq': 'str',
             'codon_type': 'str',
             'nt_pos': 'str',
             'state': 'str',
             'reads': 'float64',
             'mean_reads': 'float64',
             'norm_reads': 'float64',
             'log_reads': 'float64'}
    df.astype(dtype0, copy=False)

    return df


def read_data_file(file):
    """
    Read processed data from a file into a DataFrame
    Args:
        file: path of csv file containing exported processed data
    Returns:
        data_df: pandas DataFrame of all annotated transcripts
    Raises:

    """
    # Import the data (specify dtypes where necessary)
    dtype0 = {'transcript_id': 'str',
             'transcript_name': 'str',
             'transcript_pos': 'int64',
             'transcript_len': 'int64',
             'feature': 'str',
             'feature_pos': 'int64',
             'feature_len': 'int64',
             'seq': 'str',
             'codon_seq': 'str',
             'codon_type': 'str',
             'nt_pos': 'str',
             'state': 'str',
             'reads': 'float64',
             'mean_reads': 'float64',
             'norm_reads': 'float64',
             'log_reads': 'float64'}

    data_df = pd.read_csv(file,
                          sep='\t',
                          index_col=0,
                          compression='gzip',
                          dtype=dtype0)

    return data_df



#!/usr/bin/env python

# -----------------------------------------------------------------------------
# Hidden Markov Model
# Author: Adapted from code by Sebastian Böck <sebastian.boeck@jku.at>
# available from https://github.com/CPJKU/schimmel/blob/master/hmm.pyx
# Date: 2022.06.10
# -----------------------------------------------------------------------------

import numpy as np
cimport numpy as np
cimport cython
from cython.parallel cimport prange

cdef extern from 'math.h':
    float INFINITY


NUCLEOTIDES = ['A','C','G','T']
CODONS = [''.join(i) for i in it.product(NUCLEOTIDES, repeat = 3)]



class States():
    """States class for an HMM"""

    def __init__(self,
                 parameters,
                 start_codons=[],
                 sense_codons=[],
                 stop_codons=[]):
        """
        Construct a Transitions instance
        Args:
            states: list of state indices
            parameters: list of parameter values for
                            alpha, beta, gamma, delta, zeta
        Returns:

        Raises:

        """
        self.parameters = parameters
        self.start_codons = start_codons
        self.sense_codons = sense_codons
        self.stop_codons = stop_codons
        self.states, self.states_idx, self.altorfs_idx = \
            set_states(start_codons,
                       sense_codons,
                       stop_codons,
                       parameters)

    @property
    def num_states(self):
        """Number of states"""
        return len(self.states)

    @staticmethod
    def set_states(start_codons,
                   sense_codons,
                   stop_codons,
                   parameters):
        """
        Generate a list of all states for this model
        Args:
            start_codons: list of valid start codons
            sense_codons: list of valid sense codons
            stop_codons: list of valid stop codons
            parameters: list of parameter values for
                alpha, beta, gamma, delta, zeta
        Returns:
            states: list of state names
            states_idx: dictionary mapping state type to
                state indices (1-indexed)
            altorfs_idx: dictionary mapping more general state type to
                state indices (0-indexed)
        Raises:
            ValueError: Invalid parameter values. Make sure all parameter values
            are greater than or equal to zero.
        """
        # List parameters
        [alpha, beta, gamma, delta, zeta] = parameters

        # List states and their corresponding indices in the
        # transition and emission matrices
        states = []
        states_idx = {}

        # Create dictionary to store all indices for each type of state
        altorfs_idx = {'5UTR':[],
                       '3UTR':[],
                       'ORF':[],
                       'start':[],
                       'sense':[],
                       'stop':[],
                       'X1':[],
                       'X2':[],
                       'SCR':[],
                       'uORF':[],
                       'dORF':[]}

        # Begin counting indices
        idx = 0

        # Add UTR states
        for state in ['5UTR','3UTR']:
            states += [state]
            states_idx[state] = idx
            altorfs_idx[state] = idx
            idx += 1

        # Add start codon states
        states += [(codon + '#') for codon in sorted(start_codon_freqs)]
        n = len(start_codon_freqs)
        states_idx['start'] = list(range(idx, idx+n))
        altorfs_idx['start'] = list(range(idx, idx+n))
        altorfs_idx['ORF'] = list(range(idx, idx+n))
        idx += n

        # Add sense codon states
        states += [(codon) for state in sorted(sense_codon_freqs)]
        n = len(sense_codon_freqs)
        states_idx['sense'] = list(range(idx, idx+n))
        altorfs_idx['sense'] = list(range(idx, idx+n))
        altorfs_idx['ORF'] += list(range(idx, idx+n))
        idx += n

        # Add stop codon states
        states += [(codon + '*') for state in sorted(stop_codon_freqs)]
        n = len(stop_codon_freqs)
        states_idx['stop'] = list(range(idx, idx+n))
        altorfs_idx['stop'] = list(range(idx, idx+n))
        altorfs_idx['ORF'] += list(range(idx, idx+n))
        idx += n

        # Add PRF states
        if alpha > 0:

            # Add X1 state
            states += ['X1']
            states_idx['X1'] = idx
            altorfs_idx['X1'] = idx
            idx += 1

        if alpha > 0 and beta > 0:

            # Add X2 state
            states += ['X2']
            states_idx['X2'] = idx
            altorfs_idx['X2'] = idx
            idx += 1

        # Add SCR states
        if gamma > 0:

            # Add stop codon readthrough states
            states += [(codon + '**') for state in sorted(stop_codon_freqs)]
            n = len(stop_codon_freqs)
            states_idx['stop_SCR'] = list(range(idx, idx+n))
            altorfs_idx['SCR'] = list(range(idx, idx+n))
            idx += n

        # Add uORF and dORF states
        if delta > 0:

            # Add uORF start codon states
            states += [(codon + '#u') for state in sorted(start_codon_freqs)]
            n = len(start_codon_freqs)
            states_idx['start_uORF'] = list(range(idx, idx+n))
            altorfs_idx['start_uORF'] = list(range(idx, idx+n))
            altorfs_idx['uORF'] += list(range(idx, idx+n))
            idx += n

            # Add uORF sense codon states
            states += [(codon + 'u') for state in sorted(sense_codon_freqs)]
            n = len(sense_codon_freqs)
            states_idx['sense_uORF'] = list(range(idx, idx+n))
            altorfs_idx['uORF'] += list(range(idx, idx+n))
            idx += n

            # Add uORF stop codon states
            states += [(codon + '*u') for state in sorted(stop_codon_freqs)]
            n = len(stop_codon_freqs)
            states_idx['stop_uORF'] = list(range(idx, idx+n))
            altorfs_idx['stop_uORF'] = list(range(idx, idx+n))
            altorfs_idx['uORF'] += list(range(idx, idx+n))
            idx += n

            # Add dORF start codon states
            states += [(codon + '#d') for state in sorted(start_codon_freqs)]
            n = len(start_codon_freqs)
            states_idx['start_dORF'] = list(range(idx, idx+n))
            altorfs_idx['start_dORF'] = list(range(idx, idx+n))
            altorfs_idx['dORF'] = list(range(idx, idx+n))
            idx += n

            # Add dORF sense codon states
            states += [(codon + 'd') for state in sorted(sense_codon_freqs)]
            n = len(sense_codon_freqs)
            states_idx['sense_dORF'] = list(range(idx, idx+n))
            altorfs_idx['dORF'] += list(range(idx, idx+n))
            idx += n

            # Add dORF stop codon states
            states += [(codon + '*d') for state in sorted(stop_codon_freqs)]
            n = len(stop_codon_freqs)
            states_idx['stop_dORF'] = list(range(idx, idx+n))
            altorfs_idx['stop_dORF'] = list(range(idx, idx+n))
            altorfs_idx['dORF'] += list(range(idx, idx+n))
            idx += n

        return states, states_idx, altorfs_idx



class Transitions():
    """Transition probabilities class for an HMM"""

    def __init__(self,
                 states,
                 pointers,
                 probabilities):
        """
        Construct a Transitions instance
        Args:
            states: list of state indices
            pointers: states transitioning to state s are stored in
                states[pointers[s]:pointers[s+1]]
            probabilities: corresponding transition probabilities are stored in
                probabilities[pointers[s]:pointers[s+1]]
        Returns:

        Raises:

        """
        self.states = states
        self.pointers = pointers
        self.probabilities = probabilities

    @property
    def num_transitions(self):
        """Number of transitions"""
        return len(self.probabilities)

    @property
    def log_probabilities(self):
        """Transition log probabilities"""
        return np.log(self.probabilities)

    @staticmethod
    def make_sparse(dense_transitions):
        """
        Generate a sparse representation of transition probabilities
        Allows for fast and parallel Viterbi decoding
        Args:
            dense_transitions: numpy array of transition matrix probabilities,
                where each value is the probability of transitioning from state
                row into state col
        Returns:
            states: list of state indices
            pointers: states transitioning to state s are stored in
                indices[pointers[s]:pointers[s+1]]
            probabilities: corresponding transition probabilities are stored in
                probabilities[pointers[s]:pointers[s+1]]
        Raises:
            ValueError: Invalid transition probability matrix
        """
        from scipy.sparse import csr_matrix

        # Check that transition probabilities leaving each state sum to 1
        if not np.allclose(np.sum(dense_transitions(axis=0))), len(dense_transitions)):
            raise ValueError('Invalid transition probability matrix. \
                            Transitions leaving each state do not sum to 1.')

        # Convert transitions to a sparse CSR matrix
        # Transpose so each value is the probability of
        # transitioning from state col into state row
        transitions = csr_matrix(dense_transitions.T)

        # Extract the states, pointers, and probabilities
        # defining the sparse matrix
        states = transitions.states.astype(np.uint32)
        pointers = transitions.indptr.astype(np.uint32)
        probabilities = transitions.data.astype(dtype=np.float)

        return states, pointers, probabilities

    @classmethod
    def from_dense(cls, dense_transitions):
        """
        Create a Transitions instance from a transitions matrix
        Args:
            dense_transitions: numpy array of transition matrix probabilities,
                where each value is the probability of transitioning from state
                row into state col
        Returns:
            transitions: Transitions instance of the transition matrix
        Raises:

        """
        # Generate a sparse representation of the transitions matrix
        transitions = cls.make_sparse(dense_transitions)

        # Return a new Transitions instance
        return cls(*transitions)


class ORFeusTransitions(Transitions):

    def __init__(self,
                 data_df,
                 model_states,
                 len_5UTR=None,
                 len_3UTR=None,
                 len_orf=None,
                 len_uorf=None,
                 len_dorf=None):

        # Generate the transitions matrix as a dense array
        dense_transitions = set_probabilities(data_df,
                                              model_states,
                                              len_5UTR=None,
                                              len_3UTR=None,
                                              len_orf=None,
                                              len_uorf=None,
                                              len_dorf=None)

        # Convert the transitions matrix to a sparse array
        sparse_transitions = self.from_dense(dense_transitions)

        # Instantiate
        super(ORFeusTransitions, self).__init__(*sparse_transitions)

    @staticmethod
    def normalize(d):
        """Normalize the values of a dictionary
        Args:
            d: dictionary with numerical values
        Returns:
            normalized_d: dictionary with values normalized so they sum to 1
        """
        factor = 1.0/sum(d.values())
        normalized_d = {k: v*factor for k, v in d.items()}
        return normalized_d

    def get_codon(self, i):
        """Get the codon sequence for this state
        Args:
            i: index of state
        Returns:
            codon: codon sequence
        """
        state = self.states[i]   # Label for this state (e.g. 'ATG1#')
        codon = state[:3]        # Sequence of this codon
        return codon

    def get_nt(self, i):
        """Get the subcodon nucleotide position for this state
        Args:
            i: index of state
        Returns:
            nt: nucleotide position in codon (1, 2, or 3)
        """
        state = self.states[i] # Label for this state (e.g. 'ATG1#')
        nt = int(state[3])  # Subcodon nucleotide of this codon
        nt = nt-1
        return nt

    @staticmethod
    def get_feature_freqs(data_df):
        """
        Calculate the fraction of ORFs with 5'UTRs and 3'UTRs
        Args:
            data_df: pandas DataFrame with all annotation info
        Returns:
            f_5UTR: float indicating the fraction of ORFs having 5'UTRs
            f_3UTR: float indicating the fraction of ORFs having 3'UTRs
        Raises:
            ValueError: No ORFs found in dataset
        """
        # Calculate the total number of each feature (5UTR, 3UTR, gene)
        counts_5UTR = data_df[data_df['feature']=='5UTR']['gene_id'].nunique()
        counts_gene = data_df[data_df['feature']=='ORF']['gene_id'].nunique()
        counts_3UTR = data_df[data_df['feature']=='3UTR']['gene_id'].nunique()

        # Calculate the fraction of 5UTRs and 3UTRs
        f_5UTR = counts_5UTR/counts_gene  # Fraction of genes starting with 5'UTR
        f_3UTR = counts_3UTR/counts_gene  # Fraction of genes ending with 3'UTR

        # Check that gene count is positive
        if counts_gene <= 0:
            raise ValueError('No ORFs found in dataset.')

        return f_5UTR, f_3UTR

    @staticmethod
    def adjust_length(len,
                      orf=False):
        """
        Adjust the mean lengths of 5'UTRs, 3'UTRs, and ORFs
        Args:
            len: int mean length
            orf: boolean indicating whether this length corresponds to an ORF
        Returns:
            adjusted_len: int length converted from nt to codons if it an ORF
        Raises:
            ValueError: Invalid mean length for UTR or ORF
        """
        # Don't allow zero or negative lengths
        if len <= 0:
            adjusted_len = 1
            raise ValueError('Invalid mean length for UTR or ORF. \
                Length set to 1.')

        # Convert to codons and subtract the start and stop codons
        if orf:
            adjusted_len = adjusted_len/3 - 2

        return adjusted_len

    @staticmethod
    def get_feature_lengths(data_df,
                            len_5UTR=None,
                            len_3UTR=None,
                            len_orf=None,
                            len_uorf=None,
                            len_dorf=None):
        """
        Calculate the mean lengths of 5'UTRs, 3'UTRs, and ORFs
        Args:
            data_df: pandas DataFrame with all annotation info
        Returns:
            len_5UTR: (optional) float indicating the mean length of 5'UTRs
                default is to infer from the input data
            len_3UTR: (optional) float indicating the mean length of 3'UTRs
                default is to infer from the input data
            len_orf: (optional) float indicating the mean length of ORFs
                default is to infer from the input data
            len_uorf: (optional) float indicating the mean length of uORFs
                default is to set equal to len_orf
            len_dorf: (optional) float indicating the mean length of dORFs
                default is to set equal to len_orf
        Raises:

        """
        # Group by feature and gene so each feature is counted only once
        group = data_df.groupby(['feature','gene_id']).mean().reset_index()

        # Calculate the mean length of each feature that is not given
        if not len_5UTR:
            len_5UTR = group[group['feature']=='5UTR']['feature_length'].mean()
        if not len_3UTR:
            len_3UTR = group[group['feature']=='3UTR']['feature_length'].mean()
        if not len_orf:
            len_orf = group[group['feature']=='ORF']['feature_length'].mean()
        if not len_uorf:
            len_uorf = len_uorf
        if not len_dorf:
            len_dorf = len_dorf

        # Aujust the length
        len_5UTR = adjust_length(len_5UTR)
        len_3UTR = adjust_length(len_3UTR)
        len_orf = adjust_length(len_orf, orf=True)
        len_uorf = adjust_length(len_uorf, orf=True)
        len_dorf = adjust_length(len_dorf, orf=True)

        return len_5UTR, len_3UTR, len_orf, len_uorf, len_dorf

    @staticmethod
    def get_feature_codon_freqs(data_df,
                                feature,
                                codons=[]):
        """
        Calculate the frequency of each valid codon in this feature
        Args:
            data_df: pandas DataFrame with all annotation info
            feature: string representing the feature ('START','STOP', or 'ORF')
            codons: (optional) list of valid codons for this feature,
                default is to infer from the input data
        Returns:
            codon_freqs: dict mapping each codon to its frequency in this feature
            codons: list of codons for this feature
        Raises:

        """
        # Get the codon frequencies for this feature (with a pseudocount)
        if codons != []:
            codon_freqs = {}
            for codon in codons:
                codon_freqs[codon] = len(data_df[(data_df['codon_type']==feature) &
                                        (data_df['codon_seq']==codon)]) + 1
            codon_freqs = normalize(codon_freqs)
        else:
            codon_freqs = data_df[(data_df['codon_type']==feature)]
                            ['codon_seq'].value_counts(normalize=True).to_dict()
            codons = codon_freqs.keys()

        return codon_freqs, codons

    @staticmethod
    def get_codon_freqs(data_df,
                        start_codons=[],
                        sense_codons=[],
                        stop_codons=[]):
        """
        Calculate the frequency of start, sense, and stop codons
        Args:
            data_df: pandas DataFrame with all annotation info
            start_codons: (optional) list of valid start codons,
                default is to infer from the input data
            sense_codons: (optional) list of valid sense codons,
                default is to infer from the input data
            stop_codons: (optional) list of valid stop codons,
                default is to infer from the input data
        Returns:
            start_codons: list of start codons
            sense_codons: list of sense codons
            stop_codons: list of stop codons
            start_codon_freqs: dict mapping each start codon to its frequency
            stop_codons_freqs: dict mapping each stop codon to its frequency
            sense_codons_freqs: dict mapping each sense codon to its frequency
        Raises:
            ValueError: Invalid codons filtered out of dataset.
        """
        # Exclude NaN codons
        data_df = data_df[data_df['codon_seq'].notnull()]
        unfiltered_len = len(data_df)

        # Only consider codons that are valid
        # (exclude 2nt or 1nt fragments and invalid nucleotides)
        data_df = data_df[data_df['codon_seq'].isin(CODONS)]
        filtered_len = len(data_df)

        # Check if invalid codons were filtered out
        if filtered_len < unfiltered_len:
            raise ValueError('Invalid codons filtered out of dataset.')

        # Get codon frequencies for each feature (with a pseudocount)
        start_codon_freqs, start_codons =
            get_feature_codon_freqs(data_df, 'START', codons=start_codons)
        sense_codon_freqs, sense_codons =
            et_feature_codon_freqs(data_df, 'ORF', codons=sense_codons)
        stop_codon_freqs, stop_codons =
            get_feature_codon_freqs(data_df, 'STOP', codons=stop_codons)

        return start_codons, sense_codons, stop_codons

    @staticmethod
    def set_probabilities(data_df,
                          model_states,
                          len_5UTR=None,
                          len_3UTR=None,
                          len_orf=None,
                          len_uorf=None,
                          len_dorf=None):
        """
        Set the probability of transitioning from each state to each other state
        Args:
            data_df: pandas DataFrame with all annotation info
            model_states: States object contatining all state info for the
                current model instance
            len_5UTR: (optional) float indicating the mean length of 5'UTRs
                default is to infer from the input data
            len_3UTR: (optional) float indicating the mean length of 3'UTRs
                default is to infer from the input data
            len_orf: (optional) float indicating the mean length of ORFs
                default is to infer from the input data
            len_uorf: (optional) float indicating the mean length of uORFs
                default is to set equal to len_orf
            len_dorf: (optional) float indicating the mean length of dORFs
                default is to set equal to len_orf
        Returns:
            p_trans: the transition probability matrix
        Raises:

        """
        # Unpack parameters
        [alpha, beta, delta, gamma, epsilon] = model_states.parameters

        # Get frequency of 5'UTRs and 3'UTRs
        f_5UTR, f_3UTR = get_feature_freqs(data_df)

        # Get mean lengths of 5'UTRs, 3'UTRs, and ORFs
        len_5UTR, len_3UTR, len_orf, len_uorf, len_dorf =
            get_feature_lengths(data_df, len_5UTR, len_3UTR,
                len_orf, len_uorf, len_dorf)

        # Get start, sense, stop, and overlapping codon frequencies
        start_codons, sense_codons, stop_codons, \
            start_codon_freqs, sense_codon_freqs, stop_codon_freqs, \
            get_codon_freqs(data_df, model_states.start_codons,
                model_states.sense_codons, model_states.stop_codons):

        # Prepare the MxM transition matrix (add one to account for begin state)
        M = self.model_states.num_states
        p_trans = np.zeros((M+1,M+1))


        ################ UTR states

        # Prob of transitioning from begin to 5UTR
        p_trans[0][states_idx['5UTR']+1] = f_5UTR

        # Prob of transitioning from 5UTR to 5UTR
        p_trans[states_idx['5UTR']+1][states_idx['5UTR']+1] = \
            (len_5UTR-1)/len_5UTR * (1-delta)

        # Prob of transitioning from 3UTR to 3UTR
        p_trans[states_idx['3UTR']+1][states_idx['3UTR']+1] = \
            (len_3UTR-1)/len_3UTR * (1-delta) * (1-zeta)

        # Prob of transitioning from 3UTR to end
        p_trans[states_idx['3UTR']+1][0] = 1/len_3UTR * (1-delta)


        ################ Normal states

        # For each START codon row of the transition matrix, update the values
        for i in range(states_idx['start'][0]+1,
                       states_idx['start'][-1]+1, 3):
            codon = get_codon(i-1, states) # Sequence of this codon

            # Prob of transitioning from begin to nt1
            p_trans[0][i] = (1-f_5UTR) * start_codon_freqs[codon]

            # Prob of transitioning from 5UTR to nt1
            p_trans[states_idx['5UTR']+1][i] = \
                (1/len_5UTR) * start_codon_freqs[codon] * (1-delta)

            # Prob of transitioning from 3UTR to nt1 (multiple ORFs)
            p_trans[states_idx['3UTR']+1][i] = \
                (len_3UTR-1)/len_3UTR * start_codon_freqs[codon] * (1-delta) * zeta

            # Prob of transitioning from nt1 to nt2 within this codon
            p_trans[i][i+1] = 1

            # Prob of transitioning from nt2 to nt3 within this codon
            p_trans[i+1][i+2] = 1

            # Prob of transitioning from nt3 to nt1 in the next SENSE codon
            for j in range(states_idx['sense'][0]+1,
                           states_idx['sense'][-1]+1, 3):
                next_codon = get_codon(j-1, states) # Sequence of this codon
                p_trans[i+2][j] = sense_codon_freqs[next_codon]

        # For each SENSE codon row of the transition matrix, update the values
        for i in range(states_idx['sense'][0]+1,
                       states_idx['sense'][-1]+1, 3):
            codon = get_codon(i-1, states) # Sequence of this codon

            # Prob of transitioning from nt1 to nt2 within this codon
            p_trans[i][i+1] = 1

            # Prob of transitioning from nt2 to nt3 within this codon
            p_trans[i+1][i+2] = 1

            # Prob of transitioning from nt3 to nt1 in the next SENSE codon
            for j in range(states_idx['sense'][0]+1,
                           states_idx['sense'][-1]+1, 3):
                next_codon = get_codon(j-1, states) # Sequence of this codon
                p_trans[i+2][j] = \
                    (len_orf-1)/len_orf * sense_codon_freqs[next_codon] * (1-alpha)

            # Prob of transitioning from nt3 to nt1 in the next STOP codon
            for j in range(states_idx['stop'][0]+1, states_idx['stop'][-1]+1, 3):
                next_codon = get_codon(j-1, states) # Sequence of this codon
                p_trans[i+2][j] = \
                    1/len_orf * stop_codon_freqs[next_codon] * (1-gamma)

            # Prob of transitioning from nt3 to nt1 in the next SCR codon
            if gamma > 0:
                for j in range(states_idx['stop_SCR'][0]+1,
                               states_idx['stop_SCR'][-1]+1, 3):
                    next_codon = get_codon(j-1, states) # Sequence of this codon
                    p_trans[i+2][j] = \
                        1/len_orf * stop_codon_freqs[next_codon] * gamma

            # Prob of transitioning from nt3 to X1
            if alpha > 0:
                p_trans[i+2][states_idx['X1']+1] = \
                    (len_orf-1)/len_orf * alpha

            # Prob of transitioning from X1 to nt1 in this codon
            if alpha > 0:
                p_trans[states_idx['X1']+1][i] = \
                    sense_codon_freqs[codon] * (1-beta)

            # Prob of transitioning from X2 to nt1 in this codon
            if beta > 0:
                p_trans[states_idx['X2']+1][i] = sense_codon_freqs[codon]

        # For each STOP codon row of the transition matrix, update the values
        for i in range(states_idx['stop'][0]+1,
                       states_idx['stop'][-1]+1, 3):
            codon = get_codon(i-1, states) # Sequence of this codon

            # Prob of transitioning from nt1 to nt2 within this codon
            p_trans[i][i+1] = 1

            # Prob of transitioning from nt2 to nt3 within this codon
            p_trans[i+1][i+2] = 1

            # Prob of transitioning from nt3 to 3UTR
            p_trans[i+2][states_idx['3UTR']+1] = f_3UTR

            # Prob of transitioning from nt3 to end
            p_trans[i+2][0] = (1-f_3UTR)


        ################ SCR states
        # For each SCR codon row of the transition matrix, update the values
        if gamma > 0:
            for i in range(states_idx['stop_SCR'][0]+1,
                           states_idx['stop_SCR'][-1]+1, 3):
                codon = get_codon(i-1, states) # Sequence of this codon

                # Prob of transitioning from nt1 to nt2 within this codon
                p_trans[i][i+1] = 1

                # Prob of transitioning from nt2 to nt3 within this codon
                p_trans[i+1][i+2] = 1

                # Prob of transitioning from nt3 to nt1 in the next SENSE codon
                for j in range(states_idx['sense'][0]+1,
                               states_idx['sense'][-1]+1, 3):
                    next_codon = get_codon(j-1, states) # Sequence of this codon
                    p_trans[i+2][j] = sense_codon_freqs[next_codon]


        ################ PRF states
        if alpha > 0 and beta > 0:
            # Prob of transitioning from X1 to X2
            p_trans[states_idx['X1']+1][states_idx['X2']+1] = beta


        ################ uORF states
        if delta > 0:
            # For each START codon row of the transition matrix, update the values
            for i in range(states_idx['start_uORF'][0]+1, \
                           states_idx['start_uORF'][-1]+1, 3):
                codon = get_codon(i-1, states) # Sequence of this codon

                # Prob of transitioning from 5UTR to nt1
                p_trans[states_idx['5UTR']+1][i] = \
                    start_codon_freqs[codon] * delta

                # Prob of transitioning from nt1 to nt2 within this codon
                p_trans[i][i+1] = 1

                # Prob of transitioning from nt2 to nt3 within this codon
                p_trans[i+1][i+2] = 1

                # Prob of transitioning from nt3 to nt1 in the next SENSE codon
                for j in range(states_idx['sense_uORF'][0]+1, \
                               states_idx['sense_uORF'][-1]+1, 3):
                    next_codon = get_codon(j-1, states) # Sequence of this codon
                    p_trans[i+2][j] = sense_codon_freqs[next_codon]

            # For each SENSE codon row of the transition matrix, update the values
            for i in range(states_idx['sense_uORF'][0]+1, \
                           states_idx['sense_uORF'][-1]+1, 3):
                codon = get_codon(i-1, states) # Sequence of this codon

                # Prob of transitioning from nt1 to nt2 within this codon
                p_trans[i][i+1] = 1

                # Prob of transitioning from nt2 to nt3 within this codon
                p_trans[i+1][i+2] = 1

                # Prob of transitioning from nt3 to nt1 in the next SENSE codon
                for j in range(states_idx['sense_uORF'][0]+1, \
                               states_idx['sense_uORF'][-1]+1, 3):
                    next_codon = get_codon(j-1, states) # Sequence of this codon
                    p_trans[i+2][j] = \
                        (len_uorf-1)/len_uorf * sense_codon_freqs[next_codon]

                # Prob of transitioning from nt3 to nt1 in the next STOP codon
                for j in range(states_idx['stop_uORF'][0]+1, \
                               states_idx['stop_uORF'][-1]+1, 3):
                    next_codon = get_codon(j-1, states) # Sequence of this codon
                    p_trans[i+2][j] = \
                        1/len_uorf * stop_codon_freqs[next_codon]

            # For each STOP codon row of the transition matrix, update the values
            for i in range(states_idx['stop_uORF'][0]+1, \
                           states_idx['stop_uORF'][-1]+1, 3):
                codon = get_codon(i-1, states) # Sequence of this codon

                # Prob of transitioning from nt1 to nt2 within this codon
                p_trans[i][i+1] = 1

                # Prob of transitioning from nt2 to nt3 within this codon
                p_trans[i+1][i+2] = 1

                # Prob of transitioning from nt3 to 5UTR
                p_trans[i+2][states_idx['5UTR']+1] = 1


        ################ dORF states
        if delta > 0:
            # For each START codon row of the transition matrix, update the values
            for i in range(states_idx['start_dORF'][0]+1, \
                           states_idx['start_dORF'][-1]+1, 3):
                codon = get_codon(i-1, states) # Sequence of this codon

                # Prob of transitioning from 3UTR to nt1
                p_trans[states_idx['3UTR']+1][i] = \
                    start_codon_freqs[codon] * delta

                # Prob of transitioning from nt1 to nt2 within this codon
                p_trans[i][i+1] = 1

                # Prob of transitioning from nt2 to nt3 within this codon
                p_trans[i+1][i+2] = 1

                # Prob of transitioning from nt3 to nt1 in the next SENSE codon
                for j in range(states_idx['sense_dORF'][0]+1, \
                               states_idx['sense_dORF'][-1]+1, 3):
                    next_codon = get_codon(j-1, states) # Sequence of this codon
                    p_trans[i+2][j] = sense_codon_freqs[next_codon]

            # For each SENSE codon row of the transition matrix, update the values
            for i in range(states_idx['sense_dORF'][0]+1, \
                           states_idx['sense_dORF'][-1]+1, 3):
                codon = get_codon(i-1, states) # Sequence of this codon

                # Prob of transitioning from nt1 to nt2 within this codon
                p_trans[i][i+1] = 1

                # Prob of transitioning from nt2 to nt3 within this codon
                p_trans[i+1][i+2] = 1

                # Prob of transitioning from nt3 to nt1 in the next SENSE codon
                for j in range(states_idx['sense_dORF'][0]+1, \
                               states_idx['sense_dORF'][-1]+1, 3):
                    next_codon = get_codon(j-1, states) # Sequence of this codon
                    p_trans[i+2][j] = \
                        (len_dorf-1)/len_dorf * sense_codon_freqs[next_codon]

                # Prob of transitioning from nt3 to nt1 in the next STOP codon
                for j in range(states_idx['stop_dORF'][0]+1,
                               states_idx['stop_dORF'][-1]+1, 3):
                    next_codon = get_codon(j-1, states) # Sequence of this codon
                    p_trans[i+2][j] = 1/len_dorf * stop_codon_freqs[next_codon]

            # For each STOP codon row of the transition matrix, update the values
            for i in range(states_idx['stop_dORF'][0]+1, \
                           states_idx['stop_dORF'][-1]+1, 3):
                codon = get_codon(i-1, states) # Sequence of this codon

                # Prob of transitioning from nt1 to nt2 within this codon
                p_trans[i][i+1] = 1

                # Prob of transitioning from nt2 to nt3 within this codon
                p_trans[i+1][i+2] = 1

                # Prob of transitioning from nt3 to 3UTR
                p_trans[i+2][states_idx['3UTR']+1] = 1

        return p_trans


        def transition_probs_rev(self, model_states):
            """Calculate the the reverse transition probabilities for generating a
                sequence backwards
            Args:
                model_states: States object contatining all state info for the
                    current model instance
            Returns:
                p_trans_rev: the reverse transition probability matrix
            Raises:
                ValueError: Could not generate reverse transition matrix
            """
            # Calculate the stationary distribution (pi)

            try:
                # Calculate the eigenvector for the transpose of the transition probs
                w, v = np.linalg.eig(self.dense_transitions.T)

                # Find the first eigenvalue that is approximately 1
                w1 = int(np.where(math.isclose(w, 1, abs_tol=0.0001))[0])

                # Get the eigenvector corresponding to an eigenvalues of 1
                pi = np.real(v)[:,w1]

                # Normalize the eigenvector
                pi = pi/pi.sum()
            except:
                raise ValueError('Could not generate reverse transition matrix.')

            # Calcualte the backward transition probs
            M = model_states.num_states
            p_trans_rev = np.zeros([M,M])
            for j in range(M):
                for k in range(M):
                    p_trans_rev[j,k] = pi[k] / pi[j] * p_trans[k,j]
            p_trans_rev = p_trans_rev/p_trans_rev.sum(axis=1, keepdims=True)

            return p_trans_rev



class Emissions():
    """Emission probabilities class for an HMM"""

    def __init__(self,
                 pointers,
                 nt_probabilities,
                 rho_probabilities,
                 riboseq_ranges):
        """
        Construct an Emissions instance
        Args:
            pointers: pointers from each state to the correct emission
                probabilities column
            probabilities: corresponding emissions probabilities are stored in
                probabilities[pointers[:,s]]
        Returns:

        Raises:

        """
        self.pointers = pointers
        self.nt_probabilities = nt_probabilities
        self.rho_probabilities = rho_probabilities
        self.riboseq_ranges = riboseq_ranges


class ORFeusEmissions(Emissions):

    def __init__(self,
                 data_df,
                 model_states,
                 num_bins,
                 min_coverage=-1,
                 max_coverage=np.inf,
                 pool=False,
                 fit=False):

        # Generate a list of state pointers (indices)
        M = self.model_states.num_states
        pointers = np.arange(M, dtype=np.uint32)

        # Calculate riboseq and nt emission probabilities for each state
        riboseq_ranges = riboseq_bins(data_df, num_bins)
        rho_probabilities = set_riboseq_probabilities(data_df,
                                                      model_states,
                                                      riboseq_ranges,
                                                      min_coverage,
                                                      max_coverage,
                                                      pool,
                                                      fit)
        nt_probabilities = set_nt_probabilities(model_states)

        # Instantiate
        super(ORFeusEmissions, self).__init__(pointers,
                                              nt_probabilities,
                                              rho_probabilities,
                                              riboseq_ranges)

    @property
    def num_bins(self):
        """Number of bins"""
        return self.num_bins

    @staticmethod
    def riboseq_bins(data_df, num_bins):
        """
        Fit a distribution to the observed riboseq values for each state
        Args:
            data_df: pandas DataFrame with all riboseq info
            num_bins: int number of bins to group the riboseq data into
        Returns:
            riboseq_ranges: array of riboseq emission bin ranges

        """
        # List of ranges for discretizing riboseq observations
        min_reads = np.nanmin(data_df['norm_reads'])
        max_reads = np.nanmax(data_df['norm_reads'])

        # Create R-1 discrete bins for non-zero values
        riboseq_ranges = np.linspace(min_reads, max_reads, num=num_bins)

        # Add a bin for zeros (dummy lower boundary)
        riboseq_ranges[0] = 0
        riboseq_ranges = np.insert(riboseq_ranges, 0, -.0001) # Add a bin for zeros (dummy lower boundary)

        return riboseq_ranges

    @staticmethod
    def fit_riboseq_emissions(df,
                              states,
                              states_idx,
                              riboseq_ranges,
                              pool=False):
        """
        Fit a distribution to the frequency of observed riboseq values for
            each state
        Args:
            df: pandas DataFrame with all riboseq info
            states: list of state names
            states_idx: dictionary mapping state type to
                state indices (1-indexed)
            riboseq_ranges: array of riboseq emission bin ranges
            pool: boolean indicating whether to pool all codons of a given type
                (start, stop, sense) when calculating riboseq emissions
        Returns:
            p_emit: the riboseq emission probability matrix

        """
        # Isolate state labels
        df['nt_pos'].fillna(-1, inplace=True)
        df_states = df[['state','codon_type','nt_pos']].drop_duplicates()

        # Update SENSE states (aggregate of all SENSE codon states if pool)
        if pool:
            df_grouped = df.groupby(['codon_type','nt_pos']) \
                           .agg({'log_reads': lambda x: x.dropna().tolist(),
                                 'norm_reads': lambda x: x.dropna().tolist()})
            df_grouped = df_states.join(df_grouped,
                                        on=['codon_type','nt_pos'],
                                        how='left')

        # Update START and STOP states
        # (aggregate all START or STOP codon states)
        else:
            df_grouped1 = df.groupby(['state']) \
                            .agg({'log_reads': lambda x: x.dropna().tolist(),
                                  'norm_reads': lambda x: x.dropna().tolist()})
            df_grouped2 = df.groupby(['codon_type','nt_pos']) \
                            .agg({'log_reads': lambda x: x.dropna().tolist(),
                                  'norm_reads': lambda x: x.dropna().tolist()})
            df_grouped1 = df_states.join(df_grouped1,
                                         on='state',
                                         how='left')
            df_grouped2 = df_states.join(df_grouped2,
                                         on=['codon_type','nt_pos'],
                                         how='left')

            df_grouped1 = df_grouped1[~df_grouped1['codon_type'] \
                                     .isin(['START','STOP'])]
            df_grouped2 = df_grouped2[df_grouped2['codon_type'] \
                                     .isin(['START','STOP'])]
            df_grouped = pd.concat([df_grouped1, df_grouped2])

        # Fit a log normal distribution to the nonzero reads
        df_grouped['p'] = df_grouped.apply(lambda row: \
                                     ss.norm.fit(row['log_reads']),
                                     axis=1)

        # Calculate the the fraction of zero reads
        df_grouped['q'] = df_grouped.apply(lambda row: \
                                     float((len(row['norm_reads']) -
                                     np.count_nonzero(row['norm_reads'])) /
                                     len(row['norm_reads'])),
                                     axis=1)

        # Make sure all states are in order
        df_grouped.set_index('state', inplace=True)
        df_grouped = df_grouped.reindex(states)

        # Keep only these parameters for each state
        p = df_grouped['p'].tolist()
        q = df_grouped['q'].tolist()

        # Set representative emission values (mean each bin greater than zero)
        emit = np.mean([riboseq_ranges[:-1], riboseq_ranges[1:]], axis=0)

        # Discretize the emissions
        M = len(states)
        R = len(riboseq_ranges)-1
        p_emit = np.zeros([M,R])

        # Skip altORF states for now
        for i in (states_idx['5UTR'] + states_idx['ORF'] + states_idx['3UTR']):

            # Add the zero emissions
            if np.any(np.isnan(q[i])):
                p_emit[i,0] = None
            else:
                p_emit[i,0] = q[i]

            # Add the nonzero emissions
            for j in range(len(emit)-1):
                if np.any(np.isnan(p[i])):
                    p_emit[i,j+1] = None
                else:
                    p_emit[i,j+1] = ss.norm(*p[i]).pdf(emit[j]) * (1-q[i])

        return p_emit

    @staticmethod
    def nofit_riboseq_emissions(df,
                                states,
                                states_idx,
                                riboseq_ranges,
                                pool=False):
        """
        Calculate the frequency observed riboseq values for each state
        Args:
            df: pandas DataFrame with all riboseq info
            states: list of state names
            states_idx: dictionary mapping state type to
                state indices (1-indexed)
            riboseq_ranges: array of riboseq emission bin ranges
            pool: boolean indicating whether to pool all codons of a given type
                (start, stop, sense) when calculating riboseq emissions
        Returns:
            p_emit: the riboseq emission probability matrix

        """
        # Group the log riboseq reads by range
        # (include the upper limit of the range)
        df['riboseq_ranges'] = pd.cut(df['norm_reads'],
                                        riboseq_ranges,
                                        right=True)

        # Get an index of all states and log read ranges
        idx = pd.MultiIndex \
                .from_product([states, df['riboseq_ranges'].cat.categories],
                              names=['state', 'riboseq_ranges'])

        # For each state, count the reads in each range
        df = df.groupby(['state', 'riboseq_ranges']) \
               .size()

        # Reset the row names to make sure ALL log reads ranges are included
        # (fill in zeros for any missing ranges)
        df = df.reindex(index=idx, fill_value=0) \
               .reset_index()
        df.columns = ['state', 'riboseq_ranges', 'counts'] # Rename columns

        # Pivot to a MxR matrix of counts for each state
        df = df.pivot(index='state',
                      columns='riboseq_ranges',
                      values='counts')

        # Reorder the states
        df = df.reindex(states)

        # Emission prob is the frequency of counts in each state
        p_emit = df.to_numpy()

        # Update SENSE states (mean of all SENSE codon states if pool)
        if pool:
            for nt in [0,1,2]:
                sum_sense = \
                    p_emit[nt+states_idx['sense'][0]:states_idx['sense'][-1]:3] \
                    .sum(axis=0)
                p_emit[nt+states_idx['sense'][0]:states_idx['sense'][-1]:3] \
                    = sum_sense

        # Normalize to make sure the probs for each state sum to 1
        p_emit = p_emit / p_emit.sum(axis=1, keepdims=True)

        return p_emit

    @staticmethod
    def set_riboseq_probabilities(data_df,
                                  model_states,
                                  riboseq_ranges,
                                  min_coverage=-1,
                                  max_coverage=np.inf,
                                  pool=False,
                                  fit=False):
        """
        Calculate the probability of observing observing each range of riboseq
            values in each state
        Args:
            data_df: pandas DataFrame with all riboseq info
            model_states: States object contatining all state info for the
                current model instance
            riboseq_ranges: array of riboseq emission bin ranges
            min_coverage: (optional) int minimum mean reads per transcript
            max_coverage: (optional) int maximum mean reads per transcript
            pool: boolean indicating whether to pool all codons of a given type
                (start, stop, sense) when calculating riboseq emissions
            fit: boolean indicating whether to fit a log-normal distribution to
                the observed non-zero values when calculating riboseq emissions
        Returns:
            p_emit: the riboseq emission probability matrix
        Raises:

        """
        # Filter down to states that are in the list of states
        df = data_df[data_df['state'].isin(model_states.states)]

        # Filter down to transcripts above min coverage and below max coverage
        df = df[(df['mean_reads']>=min_coverage) & (df['mean_reads']<=max_coverage)]

        # Calculate the emission probabilities from the observed riboseq data
        # for all canonical states
        if fit:
            p_emit = fit_riboseq_emissions(df,
                                           model_states.states,
                                           model_states.states_idx,
                                           riboseq_ranges,
                                           pool)
        else:
            p_emit = nofit_riboseq_emissions(df,
                                           model_states.states,
                                           model_states.states_idx,
                                           riboseq_ranges,
                                           pool)

        # Update START states (mean of all START codon states)
        for nt in [0,1,2]:
            sum_start = p_emit[nt+states_idx['start'][0]:states_idx['start'][-1]:3] \
                        .sum(axis=0)
            p_emit[nt+states_idx['start'][0]:states_idx['start'][1]:3] \
                    = sum_start
            if 'start_uORF' in states_idx:
                p_emit[nt+states_idx['start_uORF'][0]:states_idx['start_uORF'][-1]:3] \
                    = sum_start
            if 'start_dORF' in states_idx:
                p_emit[nt+states_idx['start_dORF'][0]:states_idx['start_dORF'][-1]:3] \
                    = sum_start

        # Update STOP states (mean of all STOP codon states)
        for nt in [0,1,2]:
            sum_stop = p_emit[nt+states_idx['stop'][0]:states_idx['stop'][-1]:3] \
                       .sum(axis=0)
            p_emit[nt+states_idx['stop'][0]:states_idx['stop'][-1]:3] \
                    = sum_stop
            if 'stop_uORF' in states_idx:
                p_emit[nt+states_idx['stop_uORF'][0]:states_idx['stop_uORF'][-1]:3] \
                    = sum_stop
            if 'stop_dORF' in states_idx:
                p_emit[nt+states_idx['stop_dORF'][0]:states_idx['stop_dORF'][-1]:3] \
                    = sum_stop
            if 'stop_SCR' in states_idx:
                p_emit[nt+states_idx['stop_SCR'][0]:states_idx['stop_SCR'][-1]:3] \
                    = sum_stop

        # Add uORF and dORF states
        if 'sense_uORF' in states_idx:
            p_emit[states_idx['sense_uORF'][0]:states_idx['sense_uORF'][-1]] = \
                p_emit[states_idx['sense'][0]:states_idx['sense'][-1]]
        if 'sense_dORF' in states_idx:
            p_emit[states_idx['sense_dORF'][0]:states_idx['sense_dORF'][-1]] = \
            p_emit[states_idx['sense'][0]:states_idx['sense'][-1]]

        # Add PRF states
        mean_orf = p_emit[states_idx['sense'][0]:states_idx['sense'][-1]].mean(axis=0)
        if 'X1' in states:
            p_emit[states_idx['X1']] = mean_orf # X1
            if 'X2' in states:
                p_emit[states_idx['X2']] = mean_orf # X2

        # Replace missing states
        for row in np.where(~p_emit.any(axis=1))[0]:
            print('Emissions missing for ' + states[row])
            p_emit[row] = mean_orf

        # Add a pseudocount to all bins so there are no zero prob emission values
        p_emit += 1e-10

        # Normalize to make sure the probs for each state sum to 1
        p_emit = p_emit / p_emit.sum(axis=1, keepdims=True)

        return p_emit

    @staticmethod
    def set_nt_probabilities(model_states):
        """
        Calculate the probability of observing each nucleotide in each state
        Args:
            model_states: States object contatining all state info for the
                current model instance
        Returns:
            p_emit: the nucleotide emission probability matrix,
                i.e. if the state is ATG1, the prob of A is 1 and the prob of all
                other nucleotides is 0, and if the state is ATG2, the prob of T is
                1 and the prob of all other nucleotides is 0 (note that this array
                has one row per nucleotide, in the same order as the NUCLEOTIDES
                array)
        """
        # Prepare the nucleotide emission probabilities matrix
        M = self.model_states.num_states
        N = len(NUCLEOTIDES)
        p_emit = np.zeros((M,N)) # MxN matrix

        # Get a list of noncodon states (can be any nt)
        noncodon_states = [model_states.states_idx['5UTR'],
                           model_states.states_idx['3UTR']]
        if 'X1' in states:
            noncodon_states.append(model_states.states_idx['X1'])
        if 'X2' in states:
            noncodon_states.append(model_states.states_idx['X2'])

        # Iterate through row for each possible observed nt
        for i in range(N):

            # For each possible state
            for s in range(M):

                if s in noncodon_states:
                    # Any nt could be observed, so set value to 1/4
                    p_emit[s, i] = 1/4

                else:
                    # Only the nt at this position in this codon can be observed
                    codon = get_codon(s, model_states.states)
                    nt_pos = get_nt(s, model_states.states)

                    # If this nt is in this state, change the value to 1
                    if NUCLEOTIDES[i] == codon[nt_pos]:
                        p_emit[s, i] = 1

        return p_emit

    @staticmethod
    def logdotexp(A, B):
        """
        Code inspiration: https://stackoverflow.com/questions/23630277/numerically-stable-way-to-multiply-log-probability-matrices-in-numpy
        Given matrices A (one-hot encoded) and B (log space), calculate the
        dot product computed in logspace using the logsumexp function such that
        the log values of A are multiplied by B and then summed (in linear space)
        Args:
            A: numpy array (one-hot encoded)
            B: numpy array (log space)
        Returns:
            C: numpy array resulting from the dot product of A and B

        """
        max_B = np.max(A)
        C = np.log(np.dot(A, np.exp(B - max_B)))
        C += max_B

        return C

    @staticmethod
    def one_hot_nt(nt_seq):
        """
        Generate the one-hot encoding for this nucleotide sequence
        Args:
            nt_seq: the nucleotide sequence
        Returns:
            one_hot_nt: the one-hot-encoded matrix for this nucleotide sequence

        """
        # Length of sequence
        L = len(nt_seq)

        # Number of nucleotides
        N = len(NUCLEOTIDES)

        # Convert the sequence to numbers (A=0, C=1, G=2, T=3)
        one_hot_nt = np.zeros((L,N))
        for i,nt in enumerate(nt_seq):
            if nt in NUCLEOTIDES:
                one_hot_nt[i, NUCLEOTIDES.index(nt)] = 1
            elif nt=='N': # Any nucleotide
                one_hot_nt[i, :] = np.ones(N)
            elif nt=='R': # Purine
                one_hot_nt[i, NUCLEOTIDES.index('A')] = 1
                one_hot_nt[i, NUCLEOTIDES.index('G')] = 1
            elif nt=='Y': # Pyrimidine
                one_hot_nt[i, NUCLEOTIDES.index('C')] = 1
                one_hot_nt[i, NUCLEOTIDES.index('T')] = 1
            else:
                print('Unexpected character in sequence: %s' % i)

        return one_hot_nt

    @staticmethod
    def one_hot_riboseq(rho_seq, riboseq_ranges):
        """
        Generate the one-hot encoding for this rho sequence
        Args:
            rho_seq: the rho sequence
        Returns:
            one_hot_rho: the one-hot-encoded matrix for this rho sequence

        """
        # Length of sequence
        L = len(rho_seq)

        # Number of discretized riboseq ranges
        R = len(riboseq_ranges)-1

        # Convert the riboseq sequence to discretized ranges
        # Include the upper limit of the range and shift to be zero-indexed
        discrete_rho = np.digitize(rho_seq, riboseq_ranges, right=True) - 1

        # For anything beyond the max observed value,
        # assign it to the uppermost bin
        discrete_rho[discrete_rho >= R] = R-1

        # Create a one-hot encoding for the sequence
        one_hot_rho = np.eye(R)[np.array(discrete_rho)]

        return one_hot_rho

    def log_probabilities(self, nt_seq, rho_seq):
        """
        Generate the log emission probabilities at each position of a sequence
        Note: Returns the emission probabilities at each sequence position as
        an L x M matrix. This allows matrix addition down the line, since the
        forward and backward matrices are L x M
        Args:
            nt_seq: the nucleotide sequence
            rho_seq: the rho sequence
        Returns:
            one_hot_nt: the one-hot-encoded matrix for this nucleotide sequence
            one_hot_rho: the one-hot-encoded matrix for this rho sequence
            log_emit_seq: the total emission probabilities at each position for
                this sequence (combined nucleotide and rho emissions)
        """
        # Calculate the nucleotide emissions for this sequence
        one_hot_nt = one_hot_nt(nt_seq, NUCLEOTIDES)
        log_emit_nt = np.log(self.nt_probabilities)
        log_emit_nt_seq = logdotexp(one_hot_nt, log_emit_nt.T)

        # Create a one-hot encoding for the sequence
        one_hot_rho = one_hot_riboseq(rho_seq, self.riboseq_ranges)
        log_emit_rho = np.log(self.rho_probabilities)
        log_emit_rho_seq = logdotexp(one_hot_rho, log_emit_rho.T)

        # Calculate the total emissions for this sequence
        log_emit_seq = log_emit_nt_seq + log_emit_rho_seq

        return one_hot_nt, one_hot_rho, log_emit_seq



# Inline function to determine the best previous state
@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline void best_prev_state(int state,
                                 int pos,
                                 double [::1] current_viterbi,
                                 double [::1] previous_viterbi,
                                 unsigned int [:, ::1] viterbi_pointers,
                                 unsigned int [::1] emissions_pointers,
                                 double [:, ::1] log_emissions,
                                 unsigned int [::1] transitions_states,
                                 unsigned int [::1] transitions_pointers,
                                 double [::1] log_transitions) nogil:
    """
    Inline function to determine the best previous state
    Args:
        state: current state
        pos: current position in the input sequence
        current_viterbi: current viterbi variables
        previous_viterbi: previous viterbi variables
        viterbi_pointers: back tracking viterbi path pointers
        emissions_pointers: emission pointers
        log_emissions: log emission probabilities (for the input sequence)
        transitions_states: transition states
        transitions_pointers: transition pointers
        log_transitions: log transition probabilities
    Returns:

    """
    # Define variables
    cdef unsigned int prev_state, pointer
    cdef double density, transition_prob

    # Reset the current viterbi variable
    current_viterbi[state] = -INFINITY

    # Get the log emission probability for each state,
    # given the observed value at this position
    log_emission = log_emissions[pos, emissions_pointers[state]]

    # Iterate over all possible previous states
    for pointer in range(transition_pointers[state], \
                         transition_pointers[state + 1]):

        # Get the previous state
        prev_state = transitions_states[pointer]

        # Weight the previous state with the transition probability
        # and the current emission probability density
        transition_prob = previous_viterbi[prev_state] + \
                          log_transitions[pointer] + log_emission

        # If this transition probability is greater than the current one,
        # overwrite it and save the previous state in the current pointers
        if transition_prob > current_viterbi[state]:

            # Update the transition probability
            current_viterbi[state] = transition_prob

            # Update the back tracking pointers
            pointers[pos, state] = prev_state


class HiddenMarkovModel(object):
    """Hidden Markov Model"""

    def __init__(self,
                 states,
                 transitions,
                 emissions,
                 num_threads=1):
        """
        Construct a new Hidden Markov Model
        Args:
            states: states instance for this HMM
            transitions: transitions matrix instance for this HMM
            emissions: emissions matrix instance for this HMM
            num_threads: int number of threads for parallel Viterbi
        Returns:

        """
        self.states = states
        self.transitions = transitions
        self.emissions = emissions
        self.num_threads = num_threads


    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    def viterbi(self, nt_seq, rho_seq):
        """
        Determine the best path with the viterbi algorithm
        Args:
            nt_seq: the nucleotide sequence
            rho_seq: the rho sequence
        Returns:
            path: viterbi path
            log_probability: total log probability of viterbi path
        """
        cdef unsigned int L = len(nt_seq)
        cdef unsigned int M = self.states.num_states

        # Transitions
        transitions = self.transitions
        cdef unsigned int [::1] transitions_states = transitions.states
        cdef unsigned int [::1] transitions_pointers = transitions.pointers
        cdef double [::1] log_transitions = transitions.log_probabilities

        # Emissions
        emissions = self.emissions
        cdef unsigned int [::1] emissions_pointers = emissions.pointers
        cdef double [:, ::1] log_emissions =
            emissions.log_probabilities(nt_seq, rho_seq)

        # Current viterbi scores
        cdef double [::1] current_viterbi = np.empty(M, dtype=np.float)

        # Previous viterbi scores (initial transitions)
        cdef double [::1] previous_viterbi = log_transitions[:,0].copy()

        # Viterbi back-tracking pointers
        cdef unsigned int [:, ::1] viterbi_pointers = np.empty((L, M),
                                                          dtype=np.uint32)
        # Define counters
        cdef int state, pos
        cdef unsigned int prev_state, pointer, num_threads = self.num_threads
        cdef double transition_prob

        # Iterate over all positions of the input sequence
        for pos in range(L):

            # Search for the best transition (in parallel)
            if num_threads == 1:
                # For 1 thread, range() is faster than prange()
                for state in range(M):
                    best_prev_state(state,
                                    pos,
                                    current_viterbi,
                                    previous_viterbi,
                                    viterbi_pointers,
                                    emissions_pointers,
                                    log_emissions,
                                    transitions_states,
                                    transitions_pointers,
                                    log_transitions)
            else:
                for state in prange(M, nogil=True,
                                    schedule='static',
                                    num_threads=num_threads):
                    best_prev_state(state,
                                    pos,
                                    current_viterbi,
                                    previous_viterbi,
                                    viterbi_pointers,
                                    emissions_pointers,
                                    log_emissions,
                                    transitions_states,
                                    transitions_pointers,
                                    log_transitions)

            # Overwrite the old states with the current ones
            previous_viterbi[:] = current_viterbi

        # Get the final best state
        state = np.asarray(current_viterbi).argmax()

        # Set the path probability to that of the best state
        log_probability = current_viterbi[state]

        # Most probable viterbi path
        path = np.empty(L, dtype=np.uint32)

        # Track the path backwards
        for pos in range(L -1, -1, -1):
            # Save the state in the path
            path[pos] = state
            # Fetch the next previous one
            state = viterbi_pointers[pos, state]

        # Return the viterbi path and its total probability
        return path, log_probability

# Alias
HMM = HiddenMarkovModel



#!/usr/bin/env python

# -----------------------------------------------------------------------------
# Accuracy simulation
# Author: Mary Richardson
# Date: 2023.03.20
# -----------------------------------------------------------------------------

import logging
import numpy as np
import multiprocessing as mp
import matplotlib.pyplot as plt

logger = logging.getLogger(__name__)

from orfeus.hmm.simulation import simulation

colors = ['#6F6F6F','#DDDDDD','#E0BA4D','#9ABC61','#5385AA','#C55A56','#A472A5']


def accuracy_calculation(true_path,
                         pred_path):
    """
    Compares the true and predicted paths to determine the accuracy of the
        prediction
    Args:
        true_path: simulated path
        pred_path: predicted path
    Returns:
        accuracy: mean fraction of states that are correctly assigned in the
            predicted path
    Raises:

    """
    T=0
    # Check how many positions are correctly predicted
    for i in range(len(true_path)):
        if true_path[i] == pred_path[i]:
            T += 1
    accuracy = T/len(true_path)

    return accuracy


def accuracy(p_emit_nt,
             p_emit_rho,
             p_trans,
             p_trans_rev,
             log_emit_nt,
             log_emit_rho,
             log_trans,
             prev_states_options,
             next_states_options,
             nts,
             rho_ranges,
             altorfs_idx,
             coverage,
             event):
    """
    Simulates a path and corresponding sequence of nts and rho values,
    then runs Viterbi on the sequence to predict the path
    Args:
        p_emit_nt: the nucleotide emission probability matrix
        p_emit_rho: the riboseq emission probability matrix
        p_trans: the transition probability matrix
        p_trans_rev: the reverse transition probability matrix
        log_emit_nt: the nucleotide log emission probability matrix
        log_emit_rho: the riboseq log emission probability matrix
        log_trans: the log transition probability matrix
        prev_state_options: list of indices of all possible
            (nonzero probability) previous states
        next_state_options: list of indices of all possible
            (nonzero probability) next states
        nts: list of valid nucleotides
        rho_ranges: array of riboseq emission bin ranges
        altorfs_idx: dictionary mapping more general state type to
            state indices (0-indexed)
    Returns:
        accuracy: mean fraction of states that are correctly assigned in the
            predicted path
    Raises:

    """
    # Simulate a sequence
    true_path, pred_path = simulation(p_emit_nt,
                                      p_emit_rho,
                                      p_trans,
                                      p_trans_rev,
                                      log_emit_nt,
                                      log_emit_rho,
                                      log_trans,
                                      prev_states_options,
                                      next_states_options,
                                      nts,
                                      rho_ranges,
                                      altorfs_idx,
                                      coverage)

    # Calculate the accuracy of the prediction
    accuracy = accuracy_calculation(true_path, pred_path)

    return accuracy


def accuracy_simulation(parameters,
                        nts,
                        coverages,
                        iters):
    """
    Simulates sequences and compares the true path to the predicted path
        in parallel to estimate specificity of event detection
    Args:
        parameters: all model transition and emission parameters
        nts: list of valid nucleotides
        coverages: mean riboseq coverages to simulate when simulating sequences
        iters: number of sequences to simulate at each coverage value
    Returns:

    Raises:

    """
    states, states_idx, altorfs_idx, \
    p_emit_nt, p_emit_rho, rho_ranges, p_trans, p_trans_rev, \
    prev_states_options, next_states_options = parameters

    log_emit_nt = np.log(p_emit_nt)
    log_emit_rho = np.log(p_emit_rho)
    log_trans = np.log(p_trans)

    accuracy = {}

    for coverage in coverages:
        # Simulate sequences in parallel
        with mp.Pool(processes=10) as pool:
            args = [(p_emit_nt,
                     p_emit_rho,
                     p_trans,
                     p_trans_rev,
                     log_emit_nt,
                     log_emit_rho,
                     log_trans,
                     prev_states_options,
                     next_states_options,
                     nts,
                     rho_ranges,
                     altorfs_idx,
                     coverage)]
            results = pool.starmap(accuracy, args * iters)
        pool.join()
        res = np.array(results)

        acc = np.mean(res)
        accuracy[coverage] = acc
        print('%f coverage accuracy: %f' % (coverage, acc))

    return accuracy


def plot_accuracy(accuracy, file):
    """
    Plots mean riboseq coverage per transcript vs mean accuracy
    Args:
        accuracy: dict mapping mean riboseq coverage to mean accuracy
        file: output file for plot
    Returns:

    Raises:

    """
    # Plot the results
    fig, ax = plt.subplots(figsize=(6,4))
    x = accuracy.keys()
    y = accuracy.values()

    ax.scatter(x, y, color=colors[0]) # Plot in grey

    # Label the axes
    plt.ylim([0,1])
    plt.xlabel('Mean ORF coverage (footprints/nt)', fontsize=12)
    plt.ylabel('Mean accuracy', fontsize=12, rotation='vertical')
    plt.title('Viterbi path accuracy', fontsize=14, weight='bold')

    # Save the plot
    plt.tight_layout()
    plt.savefig(plot_file)



#!/usr/bin/env python

# -----------------------------------------------------------------------------
# Sensitivity simulation
# Author: Mary Richardson
# Date: 2023.03.20
# -----------------------------------------------------------------------------

import logging
import numpy as np
import multiprocessing as mp
import matplotlib.pyplot as plt

logger = logging.getLogger(__name__)

from orfeus.hmm.simulation import simulation

colors = ['#6F6F6F','#DDDDDD','#E0BA4D','#9ABC61','#5385AA','#C55A56','#A472A5']


def find_event(path,
               event_idx,
               i,
               window):
    """
    Determine whether the specified event is present at i +/- window in the
    path
    Args:
        path: state path
        event_idx: state indices of event
        i: position in path
        window: number of nucleotides +/- i to look
    Returns:
        True or False: whether the event is present
    Raises:

    """
    L = len(path)

    # Window to look for event
    window_min = max(0, i-window)
    window_max = min(L, i+window+1)

    if any(j in path[window_min:window_max] for j in event_idx):
       return True
    else: return False


def find_orf(true_path,
             pred_path,
             idx,
             start_idx,
             stop_idx,
             window):
    """
    Determine whether the specified event is present at i +/- window in the
    path
    Args:
        true_path: simulated path
        pred_path: predicted path
        idx: state indices of ORF codons
        start_idx: state indices of ORF start codons
        stop_idx: state indices of ORF stop codons
        window: number of nucleotides +/- i to look
    Returns:
        found: whether the ORF is present within the searched window
    Raises:

    """
    # Find the true site(s) of the event
    i = list(np.where(np.in1d(true_path, idx))[0])

    # Check the predicted path for the event
    found_start = find_event(pred_path, start_idx, min(i), window)
    found_stop = find_event(pred_path, stop_idx, max(i), window)

    found = found_start and found_stop

    return found


def find_recoding(true_path,
                  pred_path,
                  idx,
                  window):
    """
    Determine whether the specified event is present at i +/- window in the
    path
    Args:
        true_path: simulated path
        pred_path: predicted path
        idx: state indices of the recoding event
        window: number of nucleotides +/- i to look
    Returns:
        found: whether the recoding event is present within the searched window
    Raises:

    """
    # Find the true site(s) of the event
    i = list(np.where(np.in1d(true_path, idx))[0])

    # Check the predicted path for the event
    found = find_event(pred_path, idx, max(i), window)

    return found


def sensitivity_search(true_path,
                       pred_path,
                       altorfs_idx,
                       window,
                       event):
    """
    Compares the true and predicted paths to determine whether the simulated
    altORF event is correctly inferred within +/- the specified window of
    nucleotides
    Args:
        true_path: simulated path
        pred_path: predicted path
        altorfs_idx: dictionary mapping more general state type to
            state indices (0-indexed)
        window: number of nucleotides +/- true site to look
        event: event type to look for
            ORF = canonical ORF
            uORF = uORF only
            dORF = dORF only
            uORFdORF = uORF or dORF
            pPRF = +1 frameshift
            mPRF = -1 frameshift
            SCR = stop codon readthrough
    Returns:
        True or False: whether the event is correctly inferred within +/- the
            specified window in the predicted path
    Raises:

    """
    if event=='uORF':
        idx = altorfs_idx['uORF']
        start_idx = altorfs_idx['start_uORF']
        stop_idx = altorfs_idx['stop_uORF']
        return find_orf(true_path, pred_path, idx, start_idx, stop_idx, window)

    elif event=='dORF':
        idx = altorfs_idx['dORF']
        start_idx = altorfs_idx['start_dORF']
        stop_idx = altorfs_idx['stop_dORF']
        return find_orf(true_path, pred_path, idx, start_idx, stop_idx, window)

    elif event=='uORFdORF':
        idx = altorfs_idx['uORF'] + altorfs_idx['dORF']
        start_idx = altorfs_idx['start_uORF'] + altorfs_idx['start_dORF']
        stop_idx = altorfs_idx['stop_uORF'] + altorfs_idx['stop_dORF']
        return find_orf(true_path, pred_path, idx, start_idx, stop_idx, window)

    elif event=='pPRF':
        idx = altorfs_idx['PRF']
        return find_recoding(true_path, pred_path, idx, window)

    elif event=='mPRF':
        idx = altorfs_idx['PRF']
        return find_recoding(true_path, pred_path, idx, window)

    elif event=='SCR':
        idx = altorfs_idx['SCR']
        return find_recoding(true_path, pred_path, idx, window)

    elif event=='ORF':
        idx = altorfs_idx['start'] + altorfs_idx['sense'] + altorfs_idx['stop']
        idx_start = altorfs_idx['start']
        idx_stop = altorfs_idx['stop']
        return find_orf(true_path, pred_path, idx, start_idx, stop_idx, window)

    else:
        logger.error(('Invalid event type specified for sensitivity calculation. '
                      'Choose from: uORF, dORF, uORFdORF, pPRF, mPRF, SCR, ORF'))


def sensitivity(p_emit_nt,
                p_emit_rho,
                p_trans,
                p_trans_rev,
                log_emit_nt,
                log_emit_rho,
                log_trans,
                prev_states_options,
                next_states_options,
                nts,
                rho_ranges,
                altorfs_idx,
                window,
                coverage,
                event):
    """
    Simulates a path and corresponding sequence of nts and rho values,
    then runs Viterbi on the sequence to predict the path
    Args:
        p_emit_nt: the nucleotide emission probability matrix
        p_emit_rho: the riboseq emission probability matrix
        p_trans: the transition probability matrix
        p_trans_rev: the reverse transition probability matrix
        log_emit_nt: the nucleotide log emission probability matrix
        log_emit_rho: the riboseq log emission probability matrix
        log_trans: the log transition probability matrix
        prev_state_options: list of indices of all possible
            (nonzero probability) previous states
        next_state_options: list of indices of all possible
            (nonzero probability) next states
        nts: list of valid nucleotides
        rho_ranges: array of riboseq emission bin ranges
        altorfs_idx: dictionary mapping more general state type to
            state indices (0-indexed)
        window: number of nucleotides +/- true site to look
        coverage: riboseq coverage (mean reads/nt)
        event: event type to look for
            ORF = canonical ORF
            uORF = uORF only
            dORF = dORF only
            uORFdORF = uORF or dORF
            pPRF = +1 frameshift
            mPRF = -1 frameshift
            SCR = stop codon readthrough
    Returns:
        found: 1 if the true event was found in the predicted sequence, 0 if not
    Raises:

    """
    # Simulate a sequence
    true_path, pred_path = simulation(p_emit_nt,
                                      p_emit_rho,
                                      p_trans,
                                      p_trans_rev,
                                      log_emit_nt,
                                      log_emit_rho,
                                      log_trans,
                                      prev_states_options,
                                      next_states_options,
                                      nts,
                                      rho_ranges,
                                      altorfs_idx,
                                      coverage,
                                      event)

    # Check whether the event was correctly predicted
    found = sensitivity_search(true_path,
                               pred_path,
                               altorfs_idx,
                               window,
                               event)

    return int(found)


def sensitivity_simulation(parameters,
                           nts,
                           coverages,
                           iters,
                           window,
                           event):
    """
    Simulates sequences and compares the true path to the predicted path
        in parallel to estimate sensitivity of event detection
    Args:
        parameters: all model transition and emission parameters
        nts: list of valid nucleotides
        coverages: mean riboseq coverages to simulate when simulating sequences
        iters: number of sequences to simulate at each coverage value
        window: number of nucleotides +/- true site to look
        event: event type to look for
            ORF = canonical ORF
            uORF = uORF only
            dORF = dORF only
            uORFdORF = uORF or dORF
            pPRF = +1 frameshift
            mPRF = -1 frameshift
            SCR = stop codon readthrough
    Returns:

    Raises:

    """
    states, states_idx, altorfs_idx, \
    p_emit_nt, p_emit_rho, rho_ranges, p_trans, p_trans_rev, \
    prev_states_options, next_states_options = parameters

    log_emit_nt = np.log(p_emit_nt)
    log_emit_rho = np.log(p_emit_rho)
    log_trans = np.log(p_trans)

    result = {}

    for coverage in coverages:
        # Simulate sequences in parallel
        with mp.Pool(processes=10) as pool:
            args = [(p_emit_nt,
                     p_emit_rho,
                     p_trans,
                     p_trans_rev,
                     log_emit_nt,
                     log_emit_rho,
                     log_trans,
                     prev_states_options,
                     next_states_options,
                     nts,
                     rho_ranges,
                     altorfs_idx,
                     window,
                     coverage,
                     event)]
            results = pool.starmap(sensitivity, args * iters)
        pool.join()
        res = np.array(results)

        sens = sum(res)/len(res) # TP/(TP+FN)
        result[coverage] = sens
        print('%f coverage sensitivity: %f' % (coverage, sens))

    return result


def plot_sensitivity(sensitivity, event, file):
    """
    Plots mean riboseq coverage per transcript vs mean sensitivity
    Args:
        sensitivity: dict mapping mean riboseq coverage to mean sensitivity
        event: event type to look for
            pred_path = canonical ORF
            uORF = uORF only
            dORF = dORF only
            uORFdORF = uORF or dORF
            pPRF = +1 frameshift
            mPRF = -1 frameshift
            SCR = stop codon readthrough
        file: output file for plot
    Returns:

    Raises:

    """
    # Set plot color
    if event=='uORF' or event=='dORF' or event=='uORFdORF': c = colors[2]
    elif event=='pPRF': c = colors[4]
    elif event=='mPRF': c = colors[5]
    elif event=='SCR': c = colors[6]

    fig, ax = plt.subplots(figsize=(6,4))

    # Plot the results
    x = sensitivity.keys()
    y = sensitivity.values()
    ax.scatter(x, y, color=c)

    # Label the axes
    plt.ylim([0,1])
    plt.xlabel('Mean ORF coverage (footprints/nt)', fontsize=12)
    plt.ylabel('Sensitivity', fontsize=12, rotation='vertical')
    plt.title('Sensitivity', fontsize=14, weight='bold')

    # Save the plot
    plt.tight_layout()
    plt.savefig(file)



#!/usr/bin/env python

# -----------------------------------------------------------------------------
# Gene coverage
# Author: Mary Richardson
# Date: 2023.03.20
# -----------------------------------------------------------------------------

import logging
import matplotlib.pyplot as plt

logger = logging.getLogger(__name__)



def count_gene_coverage(data_df,
                        coverage):
    """
    Compares the true and predicted paths to determine the accuracy of the
        prediction
    Args:
        data_df: pandas DataFrame of all annotated transcripts
        coverage: mean coverage per transcript (reads/nt)
    Returns:
        count: number of transcripts with the given coverage or higher
    Raises:

    """
    # Count the number of genes with at least the specified coverage
    count = len(data_df[data_df['mean_reads']>=coverage] \
                       ['transcript_id'].unique())

    return count


def plot_gene_coverage(data_df, coverages, file):
    """
    Compares the true and predicted paths to determine the accuracy of the
        prediction
    Args:
        data_df: pandas DataFrame of all annotated transcripts
        coverages: mean coverages per transcript (reads/nt)
        file: output file for plot
    Returns:

    Raises:

    """
    # Calculate the genes for each coverage
    coverages = np.insert(coverages,0,0)
    counts = [count_gene_coverage(data_df, coverage) for coverage in coverages]

    # Plot the results
    fig, ax = plt.subplots(figsize=(6,4))
    x = coverages
    y = counts

    print('Total genes: %i' % counts[0])
    for i in range(len(counts)):
        print('%f coverage genes: %i' % (coverages[i], counts[i]))

    ax.scatter(x, y, color=colors[0]) # Plot in grey

    # Label the axes
    plt.xlabel('Mean ORF coverage (footprints/nt)', fontsize=12)
    plt.ylabel('Number of genes', fontsize=12, rotation='vertical')
    plt.title('Number of genes\nwith sufficient coverage', fontsize=14, weight='bold')

    # Save the plot
    plt.tight_layout()
    plt.savefig(file)



#!/usr/bin/env python

# -----------------------------------------------------------------------------
# WIG READER
# Author: Mary Richardson
# Date: 2022.11.02
# -----------------------------------------------------------------------------

import re
import gzip
import logging
import pandas as pd
from collections import defaultdict

logger = logging.getLogger(__name__)


READS_HEADER  = ['chrom', 'chrom_pos', 'reads']
PARAMETERS = {'step_type':'', 'chrom':'', 'start':0, 'step':0, 'span':1}

R_KEYVALUE = re.compile(r'(\s*=\s*)')
R_CHROM  = re.compile(r'(\s*chr\s*)')


def wig_to_df(file):
    """
    Open a Wiggle file and return a pandas DataFrame
    Wiggle file format: ensembl.org/info/website/upload/wig.html
    Args:
        file: path to a read counts wig file
    Returns:
        pandas DataFrame with the read counts
    Raises:
        IOError: an error occurred accessing the read counts file
        IndexError, TypeError, KeyError: an error occurred reading the
            read counts file
    """
    # Each column is a list stored as a value in this dict
    reads = defaultdict(list)

    for i, line in enumerate(lines(file)):
        for key in line.keys():
            # This key has not been seen yet, so set it to None for all
            # previous lines
            if key not in reads:
                reads[key] = [None] * i

        # Ensure this row has some value for each column
        for key in reads.keys():
            reads[key].append(line.get(key, None))

    return pd.DataFrame(reads)


def lines(file):
    # Open the read counts wig file for reading
    try:
        if file.endswith('.gz'): f = gzip.open(file,'r')
        else: f = open(file,'r')

    except IOError:
        logging.error('Could not open read counts file %s' % file)

    else:
        with f:
            try:
                for line in f:

                    # Skip header lines
                    if line.startswith('#'):
                        continue
                    elif line.startswith('track'):
                        continue

                    # Set variables from info lines
                    elif line.startswith('fixedStep'):
                        set_variables(line)
                    elif line.startswith('variableStep'):
                        set_variables(line)

                    # Parse read count data lines
                    else:
                        fields = line.rstrip().split()

                        step_type = PARAMETERS['step_type']
                        chrom = PARAMETERS['chrom']
                        start = int(PARAMETERS['start'])
                        step = int(PARAMETERS['step'])
                        span = int(PARAMETERS['span'])

                        if step_type=='fixedStep':
                            for pos in range(start,start+span):
                                reads = float(fields[0])
                                yield set_read_counts(chrom, pos, reads)
                            PARAMETERS['start'] = start + step

                        elif PARAMETERS['step_type']=='variableStep':
                            start = int(fields[0])
                            reads = float(fields[1])
                            for pos in range(start,start+span):
                                yield set_read_counts(chrom, pos, reads)

            except (IndexError, TypeError, KeyError):
                logging.warning('Could not parse %s as Wiggle file.' % file)
                raise SyntaxError


def set_variables(line):
    fields = line.rstrip().split()

    PARAMETERS['step_type'] = fields[0]

    # Field consists of "key1=value key2=value ..."
    for i, info in enumerate(fields[1:]):
        # It should be key=value
        key, _, value = re.split(R_KEYVALUE, info, 1)
        # Ignore the field if there is no value
        if value:
            PARAMETERS[key] = _get_value(value)


def set_read_counts(chrom, pos, reads):
    read_counts = {}

    read_counts['chrom'] = chrom
    read_counts['chrom_pos'] = pos
    read_counts['reads'] = reads

    return read_counts


def _get_value(value):
    if not value:
        return None

    # Return chromosome number
    if 'chr' in value:
        value = re.split(R_CHROM, value)[1]

    # These values are equivalent to None
    elif value in ['', '.', 'NA']:
        return None

    return value



#!/usr/bin/env python

# -----------------------------------------------------------------------------
# Set parameters helper functions
# Author: Mary Richardson
# Date: 2020.07.07, last update 2023.03.02
# -----------------------------------------------------------------------------



def get_codon(i, states):
    """
    Get the codon sequence for this state
    Args:
        i: index of state
        states: list of all states in the model
    Returns:
        codon: codon sequence
    """
    state = states[i] # Label for this state (e.g. 'ATG1#')
    codon = state[:3] # Sequence of this codon
    return codon


def normalize(d):
    """
    Normalize the values of a dictionary
    Args:
        d: dictionary with numerical values
    Returns:
        normalized_d: dictionary with values normalized so they sum to 1
    """
    factor = 1.0/sum(d.values())
    normalized_d = {k: v*factor for k, v in d.items()}
    return normalized_d
